# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Salesforce Integration

Core Salesforce REST API integration class.
"""

import logging
import os
from typing import Any, Dict, List, Optional

from ..base import PaginatedIntegration, IntegrationConfig, APIResponse

logger = logging.getLogger(__name__)


class SalesforceIntegration(PaginatedIntegration):
    """
    Salesforce REST API integration.

    Usage:
        async with SalesforceIntegration(
            instance_url="https://yourorg.salesforce.com",
            access_token="your_access_token",
        ) as sf:
            # Create a lead
            await sf.create("Lead", {
                "FirstName": "John",
                "LastName": "Doe",
                "Company": "Acme Corp",
            })

            # Query records
            await sf.query("SELECT Id, Name FROM Account LIMIT 10")
    """

    service_name = "salesforce"
    api_version = "v59.0"

    def __init__(
        self,
        instance_url: Optional[str] = None,
        access_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        """
        Initialize Salesforce integration.

        Args:
            instance_url: Salesforce instance URL (e.g., https://yourorg.salesforce.com)
            access_token: OAuth access token
            refresh_token: OAuth refresh token for auto-refresh
            client_id: Connected App client ID (for refresh)
            client_secret: Connected App client secret (for refresh)
        """
        self.instance_url = (
            instance_url
            or os.getenv("SALESFORCE_INSTANCE_URL")
            or "https://login.salesforce.com"
        )
        self.refresh_token = refresh_token
        self.client_id = client_id or os.getenv("SALESFORCE_CLIENT_ID")
        self.client_secret = client_secret or os.getenv("SALESFORCE_CLIENT_SECRET")

        token = access_token or os.getenv("SALESFORCE_ACCESS_TOKEN")

        base_url = f"{self.instance_url.rstrip('/')}/services/data"
        config = IntegrationConfig(
            service_name="salesforce",
            base_url=base_url,
            api_version=self.api_version,
            rate_limit_calls=100,
            rate_limit_period=60,
        )

        super().__init__(access_token=token, config=config)

    def _get_auth_header(self) -> Dict[str, str]:
        """Get Salesforce authorization header."""
        if self.access_token:
            return {"Authorization": f"Bearer {self.access_token}"}
        return {}

    # CRUD Operations

    async def create(
        self,
        sobject: str,
        data: Dict[str, Any],
    ) -> APIResponse:
        """
        Create a new record.

        Args:
            sobject: Object type (Account, Contact, Lead, etc.)
            data: Field values

        Returns:
            APIResponse with created record ID
        """
        return await self.post(f"sobjects/{sobject}", json=data)

    async def get(
        self,
        sobject: str,
        record_id: str,
        fields: Optional[List[str]] = None,
    ) -> APIResponse:
        """
        Get a record by ID.

        Args:
            sobject: Object type
            record_id: Record ID
            fields: Specific fields to retrieve

        Returns:
            APIResponse with record data
        """
        endpoint = f"sobjects/{sobject}/{record_id}"
        params = {}
        if fields:
            params["fields"] = ",".join(fields)
        return await super().get(endpoint, params=params)

    async def update(
        self,
        sobject: str,
        record_id: str,
        data: Dict[str, Any],
    ) -> APIResponse:
        """
        Update a record.

        Args:
            sobject: Object type
            record_id: Record ID
            data: Fields to update

        Returns:
            APIResponse with result
        """
        return await self.patch(f"sobjects/{sobject}/{record_id}", json=data)

    async def upsert(
        self,
        sobject: str,
        external_id_field: str,
        external_id: str,
        data: Dict[str, Any],
    ) -> APIResponse:
        """
        Upsert a record using external ID.

        Args:
            sobject: Object type
            external_id_field: External ID field name
            external_id: External ID value
            data: Field values

        Returns:
            APIResponse with result
        """
        return await self.patch(
            f"sobjects/{sobject}/{external_id_field}/{external_id}",
            json=data,
        )

    async def delete_record(
        self,
        sobject: str,
        record_id: str,
    ) -> APIResponse:
        """
        Delete a record.

        Args:
            sobject: Object type
            record_id: Record ID

        Returns:
            APIResponse with result
        """
        return await self.delete(f"sobjects/{sobject}/{record_id}")

    # Query Operations

    async def query(self, soql: str) -> APIResponse:
        """
        Execute SOQL query.

        Args:
            soql: SOQL query string

        Returns:
            APIResponse with query results
        """
        from urllib.parse import quote
        return await super().get(f"query?q={quote(soql)}")

    async def query_all(
        self,
        soql: str,
        include_deleted: bool = False,
    ) -> List[Dict[str, Any]]:
        """
        Execute SOQL query and fetch all results (handles pagination).

        Args:
            soql: SOQL query string
            include_deleted: Include deleted/archived records

        Returns:
            List of all matching records
        """
        from urllib.parse import quote

        endpoint = "queryAll" if include_deleted else "query"
        all_records = []
        next_url = f"{endpoint}?q={quote(soql)}"

        while next_url:
            response = await super().get(next_url)
            if not response.ok:
                break

            data = response.data
            records = data.get("records", [])
            all_records.extend(records)

            # Check for more records
            if data.get("done", True):
                break

            next_records_url = data.get("nextRecordsUrl")
            if next_records_url:
                # Extract path after /services/data/vXX.X/
                next_url = next_records_url.split(f"/{self.api_version}/")[-1]
            else:
                break

        return all_records

    async def search(self, sosl: str) -> APIResponse:
        """
        Execute SOSL search.

        Args:
            sosl: SOSL search string

        Returns:
            APIResponse with search results
        """
        from urllib.parse import quote
        return await super().get(f"search?q={quote(sosl)}")

    # Describe Operations

    async def describe(self, sobject: str) -> APIResponse:
        """
        Describe an object's metadata.

        Args:
            sobject: Object type

        Returns:
            APIResponse with object metadata
        """
        return await super().get(f"sobjects/{sobject}/describe")

    async def describe_global(self) -> APIResponse:
        """Get list of all available objects."""
        return await super().get("sobjects")

    # Composite Operations

    async def composite(
        self,
        requests: List[Dict[str, Any]],
        all_or_none: bool = False,
    ) -> APIResponse:
        """
        Execute composite request (multiple operations in one call).

        Args:
            requests: List of subrequests
            all_or_none: Rollback all if any fails

        Returns:
            APIResponse with results
        """
        return await self.post("composite", json={
            "allOrNone": all_or_none,
            "compositeRequest": requests,
        })

    async def composite_batch(
        self,
        requests: List[Dict[str, Any]],
        halt_on_error: bool = False,
    ) -> APIResponse:
        """
        Execute batch request (up to 25 subrequests).

        Args:
            requests: List of subrequests
            halt_on_error: Stop on first error

        Returns:
            APIResponse with results
        """
        return await self.post("composite/batch", json={
            "haltOnError": halt_on_error,
            "batchRequests": requests,
        })

    # Bulk Operations

    async def bulk_create(
        self,
        sobject: str,
        records: List[Dict[str, Any]],
    ) -> APIResponse:
        """
        Create multiple records using Composite API.

        Args:
            sobject: Object type
            records: List of records to create

        Returns:
            APIResponse with results
        """
        return await self.post(f"composite/sobjects/{sobject}", json={
            "allOrNone": False,
            "records": [{"attributes": {"type": sobject}, **r} for r in records],
        })

    async def bulk_update(
        self,
        sobject: str,
        records: List[Dict[str, Any]],
    ) -> APIResponse:
        """
        Update multiple records (each must have 'Id' field).

        Args:
            sobject: Object type
            records: List of records with Id and fields to update

        Returns:
            APIResponse with results
        """
        return await self.patch(f"composite/sobjects/{sobject}", json={
            "allOrNone": False,
            "records": [{"attributes": {"type": sobject}, **r} for r in records],
        })

    async def bulk_delete(
        self,
        record_ids: List[str],
        all_or_none: bool = False,
    ) -> APIResponse:
        """
        Delete multiple records.

        Args:
            record_ids: List of record IDs
            all_or_none: Rollback all if any fails

        Returns:
            APIResponse with results
        """
        ids_param = ",".join(record_ids)
        return await self.delete(
            "composite/sobjects",
            params={"ids": ids_param, "allOrNone": str(all_or_none).lower()},
        )

    # Convenience Methods

    async def get_account(self, account_id: str) -> APIResponse:
        """Get Account record."""
        return await self.get("Account", account_id)

    async def get_contact(self, contact_id: str) -> APIResponse:
        """Get Contact record."""
        return await self.get("Contact", contact_id)

    async def get_lead(self, lead_id: str) -> APIResponse:
        """Get Lead record."""
        return await self.get("Lead", lead_id)

    async def get_opportunity(self, opp_id: str) -> APIResponse:
        """Get Opportunity record."""
        return await self.get("Opportunity", opp_id)



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Atomic Path Operations
Join, dirname, basename, extension, normalize, and is_absolute
"""

try:
    from .join import *
except ImportError:
    pass

try:
    from .dirname import *
except ImportError:
    pass

try:
    from .basename import *
except ImportError:
    pass

try:
    from .extension import *
except ImportError:
    pass

try:
    from .normalize import *
except ImportError:
    pass

try:
    from .is_absolute import *
except ImportError:
    pass

__all__ = []



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Browser Click Module - Click an element on the page
"""
from typing import Any, Dict
from ...base import BaseModule
from ...registry import register_module
from ...schema import compose, field
from ...schema.constants import FieldGroup
from ...schema import presets


@register_module(
    module_id='browser.click',
    version='1.1.0',
    category='browser',
    tags=['browser', 'interaction', 'click', 'ssrf_protected'],
    label='Click Element',
    label_key='modules.browser.click.label',
    description='Click an element on the page',
    description_key='modules.browser.click.description',
    icon='MousePointerClick',
    color='#F0AD4E',

    # Connection types
    input_types=['page'],
    output_types=['page'],

    can_receive_from=['browser.*', 'flow.*'],
    can_connect_to=['browser.*', 'element.*', 'flow.*', 'data.*', 'string.*', 'array.*', 'object.*', 'file.*'],
    params_schema=compose(
        field("click_method", type="select",
              label="How to find the element",
              label_key="modules.browser.click.param.click_method.label",
              description="Choose the easiest way to identify the element you want to click",
              description_key="modules.browser.click.param.click_method.description",
              default="text",
              options=[
                  {"value": "text", "label": "By text on the page",
                   "label_key": "modules.browser.click.param.click_method.option.text"},
                  {"value": "button", "label": "By button / link text",
                   "label_key": "modules.browser.click.param.click_method.option.button"},
                  {"value": "id", "label": "By element ID",
                   "label_key": "modules.browser.click.param.click_method.option.id"},
                  {"value": "selector", "label": "CSS / XPath selector (advanced)",
                   "label_key": "modules.browser.click.param.click_method.option.selector"},
              ],
              group=FieldGroup.BASIC),
        field("target", type="string",
              label="What to click",
              label_key="modules.browser.click.param.target.label",
              description='e.g. "Submit", "Next Page", "Login"',
              description_key="modules.browser.click.param.target.description",
              placeholder="Submit",
              showIf={"click_method": {"$in": ["text", "button", "id"]}},
              group=FieldGroup.BASIC),
        field("selector", type="string",
              label="CSS/XPath Selector",
              label_key="schema.field.selector",
              description="CSS selector, XPath, or text selector",
              placeholder='#submit-btn, .btn-primary, //button[@type="submit"]',
              showIf={"click_method": "selector"},
              group=FieldGroup.BASIC),
        field("button", type="select",
              label="Mouse Button",
              label_key="modules.browser.click.param.button.label",
              description="Which mouse button to use for clicking",
              default="left",
              options=[
                  {"value": "left", "label": "Left"},
                  {"value": "right", "label": "Right"},
                  {"value": "middle", "label": "Middle"},
              ],
              group=FieldGroup.OPTIONS),
        field("click_count", type="integer",
              label="Click Count",
              label_key="modules.browser.click.param.click_count.label",
              description="Number of clicks (2 for double-click, 3 for triple-click)",
              default=1,
              min=1,
              max=3,
              group=FieldGroup.OPTIONS),
        field("force", type="boolean",
              label="Force Click",
              label_key="modules.browser.click.param.force.label",
              description="Force click even if element is not actionable (covered, invisible)",
              default=False,
              group=FieldGroup.ADVANCED),
        field("modifiers", type="array",
              label="Keyboard Modifiers",
              label_key="modules.browser.click.param.modifiers.label",
              description="Modifier keys to hold during click",
              required=False,
              group=FieldGroup.ADVANCED),
        presets.TIMEOUT_MS(default=30000),
    ),
    output_schema={
        'status': {'type': 'string', 'description': 'Operation status (success/error)',
                'description_key': 'modules.browser.click.output.status.description'},
        'selector': {'type': 'string', 'description': 'Selector that was used',
                'description_key': 'modules.browser.click.output.selector.description'},
        'method': {'type': 'string', 'description': 'Click method used'}
    },
    examples=[
        {
            'name': 'Click by button text',
            'params': {'click_method': 'text', 'target': 'Submit'}
        },
        {
            'name': 'Click by element ID',
            'params': {'click_method': 'id', 'target': 'login-button'}
        },
        {
            'name': 'Click with CSS selector',
            'params': {'click_method': 'selector', 'selector': '#submit-button'}
        }
    ],
    author='Flyto Team',
    license='MIT',
    timeout_ms=30000,
    required_permissions=["browser.automation"],
)
class BrowserClickModule(BaseModule):
    """Click Element Module"""

    module_name = "Click Element"
    module_description = "Click an element on the page"
    required_permission = "browser.automation"

    def validate_params(self) -> None:
        method = self.params.get('click_method', 'text')
        target = self.params.get('target', '').strip()
        raw_selector = self.params.get('selector', '').strip()

        # Backward compatibility: selector provided without click_method → selector mode
        if 'click_method' not in self.params and raw_selector and not target:
            method = 'selector'

        if method == 'selector':
            if not raw_selector:
                raise ValueError("CSS/XPath selector is required in advanced mode")
            self.selector = raw_selector
        elif method == 'id':
            if not target:
                raise ValueError("Element ID is required")
            self.selector = f'#{target.lstrip("#")}'
        elif method == 'button':
            if not target:
                raise ValueError("Button or link text is required")
            self.selector = f':is(button, a, [role="button"]):has-text("{target}")'
        else:  # text (default)
            if not target:
                raise ValueError("Text content is required")
            self.selector = f'text={target}'

        self.method = method
        self.button = self.params.get('button', 'left')
        self.click_count = self.params.get('click_count', 1)
        self.force = self.params.get('force', False)
        self.modifiers = self.params.get('modifiers', [])

    async def execute(self) -> Any:
        browser = self.context.get('browser')
        if not browser:
            raise RuntimeError("Browser not launched. Please run browser.launch first")

        # Wait for element to be visible before clicking (unless force mode)
        if not self.force:
            await browser.wait(self.selector, state='visible', timeout_ms=10000)

        page = browser.page

        click_options = {
            'button': self.button,
            'click_count': self.click_count,
            'force': self.force,
        }
        if self.modifiers:
            click_options['modifiers'] = self.modifiers

        await page.click(self.selector, **click_options)
        return {"status": "success", "selector": self.selector, "method": self.method}





# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Validation Presets
"""
from __future__ import annotations
from typing import Any, Dict, Optional
from ..builders import field
from ..constants import FieldGroup


def VALIDATE_EMAIL(
    *,
    key: str = "email",
    required: bool = True,
    label: str = "Email",
    label_key: str = "schema.field.validate_email",
) -> Dict[str, Dict[str, Any]]:
    """Email address to validate."""
    return field(
        key,
        type="string",
        label=label,
        label_key=label_key,
        required=required,
        placeholder="user@example.com",
        description='Email address to validate',
        group=FieldGroup.BASIC,
    )


def VALIDATE_URL(
    *,
    key: str = "url",
    required: bool = True,
    label: str = "URL",
    label_key: str = "schema.field.validate_url",
) -> Dict[str, Dict[str, Any]]:
    """URL to validate."""
    return field(
        key,
        type="string",
        label=label,
        label_key=label_key,
        required=required,
        placeholder="https://example.com",
        description='URL to validate',
        group=FieldGroup.BASIC,
    )


def VALIDATE_PHONE(
    *,
    key: str = "phone",
    required: bool = True,
    label: str = "Phone Number",
    label_key: str = "schema.field.validate_phone",
) -> Dict[str, Dict[str, Any]]:
    """Phone number to validate."""
    return field(
        key,
        type="string",
        label=label,
        label_key=label_key,
        required=required,
        placeholder="+1234567890",
        description='Phone number to validate',
        group=FieldGroup.BASIC,
    )


def VALIDATE_UUID(
    *,
    key: str = "uuid",
    required: bool = True,
    label: str = "UUID",
    label_key: str = "schema.field.validate_uuid",
) -> Dict[str, Dict[str, Any]]:
    """UUID to validate."""
    return field(
        key,
        type="string",
        label=label,
        label_key=label_key,
        required=required,
        placeholder="550e8400-e29b-41d4-a716-446655440000",
        description='UUID to validate',
        group=FieldGroup.BASIC,
    )


def VALIDATE_IP(
    *,
    key: str = "ip",
    required: bool = True,
    label: str = "IP Address",
    label_key: str = "schema.field.validate_ip",
) -> Dict[str, Dict[str, Any]]:
    """IP address to validate."""
    return field(
        key,
        type="string",
        label=label,
        label_key=label_key,
        required=required,
        placeholder="192.168.1.1",
        description='IP address to validate',
        group=FieldGroup.BASIC,
    )


def VALIDATE_CARD(
    *,
    key: str = "card_number",
    required: bool = True,
    label: str = "Card Number",
    label_key: str = "schema.field.validate_card",
) -> Dict[str, Dict[str, Any]]:
    """Credit card number to validate."""
    return field(
        key,
        type="string",
        label=label,
        label_key=label_key,
        required=required,
        placeholder="4111111111111111",
        description='Credit card number to validate',
        group=FieldGroup.BASIC,
    )



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Test Runner Executor

Executes workflow tests and generates reports.
"""

import asyncio
import json
import logging
import time
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import yaml

from .models import TestCase, TestReport, TestResult, TestStatus

logger = logging.getLogger(__name__)


class WorkflowTestRunner:
    """
    Executes workflow tests and generates reports.

    Usage:
        runner = WorkflowTestRunner()

        # Load tests from file
        runner.load_tests("tests/workflow_tests.yaml")

        # Run all tests
        report = await runner.run_all()

        # Run specific tests
        report = await runner.run_tests(["test_1", "test_2"])

        # Run by tags
        report = await runner.run_by_tags(["regression"])
    """

    def __init__(
        self,
        workflow_executor: Optional[Callable] = None,
        parallel: bool = False,
        max_parallel: int = 5,
    ):
        """
        Initialize test runner.

        Args:
            workflow_executor: Async function to execute workflows
                Signature: async def executor(workflow, params) -> result
            parallel: Whether to run tests in parallel
            max_parallel: Maximum parallel tests
        """
        self._tests: Dict[str, TestCase] = {}
        self._executor = workflow_executor or self._default_executor
        self._parallel = parallel
        self._max_parallel = max_parallel

    async def _default_executor(
        self,
        workflow: Dict[str, Any],
        params: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Default workflow executor"""
        try:
            from ...engine import WorkflowEngine
        except ImportError:
            from src.core.engine import WorkflowEngine

        engine = WorkflowEngine(workflow=workflow, params=params)
        result = await engine.execute()

        return {
            "ok": result.get('ok', True),
            "status": "success" if result.get('ok', True) else "error",
            "context": engine.context,
            "execution_log": engine.execution_log,
            "outputs": result.get('outputs', {}),
            "error": result.get('error'),
        }

    def add_test(self, test: TestCase) -> None:
        """Add a test case"""
        self._tests[test.name] = test

    def load_tests(self, path: Union[str, Path]) -> int:
        """
        Load tests from YAML/JSON file.

        Args:
            path: Path to test file

        Returns:
            Number of tests loaded
        """
        path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f"Test file not found: {path}")

        with open(path, 'r', encoding='utf-8') as f:
            if path.suffix in ('.yaml', '.yml'):
                data = yaml.safe_load(f)
            else:
                data = json.load(f)

        # Handle single test or list of tests
        tests = data.get('tests', [data]) if isinstance(data, dict) else data

        count = 0
        for test_data in tests:
            test = TestCase.from_dict(test_data)
            self.add_test(test)
            count += 1

        logger.info(f"Loaded {count} tests from {path}")
        return count

    def load_workflow(
        self,
        workflow_ref: Union[Dict[str, Any], str]
    ) -> Dict[str, Any]:
        """
        Load workflow from reference.

        Args:
            workflow_ref: Workflow dict or path

        Returns:
            Workflow definition
        """
        if isinstance(workflow_ref, dict):
            return workflow_ref

        path = Path(workflow_ref)
        if not path.exists():
            raise FileNotFoundError(f"Workflow not found: {path}")

        with open(path, 'r', encoding='utf-8') as f:
            if path.suffix in ('.yaml', '.yml'):
                return yaml.safe_load(f)
            else:
                return json.load(f)

    async def run_test(self, test: TestCase) -> TestResult:
        """
        Run a single test.

        Args:
            test: Test case to run

        Returns:
            Test result
        """
        started_at = datetime.now()

        # Handle skipped tests
        if test.skip:
            return TestResult(
                test_name=test.name,
                status=TestStatus.SKIPPED,
                error=test.skip_reason or "Test skipped",
                started_at=started_at,
                finished_at=datetime.now(),
            )

        try:
            # Run setup
            if test.setup:
                if asyncio.iscoroutinefunction(test.setup):
                    await test.setup()
                else:
                    test.setup()

            # Load workflow
            workflow = self.load_workflow(test.workflow)

            # Execute workflow with timeout
            start_time = time.time()
            try:
                result = await asyncio.wait_for(
                    self._executor(workflow, test.inputs),
                    timeout=test.timeout_ms / 1000,
                )
            except asyncio.TimeoutError:
                return TestResult(
                    test_name=test.name,
                    status=TestStatus.ERROR,
                    error=f"Test timed out after {test.timeout_ms}ms",
                    duration_ms=test.timeout_ms,
                    started_at=started_at,
                    finished_at=datetime.now(),
                )

            duration_ms = int((time.time() - start_time) * 1000)

            # Run assertions
            assertion_results = []
            for assertion in test.assertions:
                assertion_result = assertion.check(result)
                assertion_results.append(assertion_result)

            # Determine overall status
            all_passed = all(a.passed for a in assertion_results)
            status = TestStatus.PASSED if all_passed else TestStatus.FAILED

            return TestResult(
                test_name=test.name,
                status=status,
                assertions=assertion_results,
                duration_ms=duration_ms,
                execution_result=result,
                started_at=started_at,
                finished_at=datetime.now(),
            )

        except Exception as e:
            logger.error(f"Test {test.name} error: {e}")
            return TestResult(
                test_name=test.name,
                status=TestStatus.ERROR,
                error=str(e),
                started_at=started_at,
                finished_at=datetime.now(),
            )

        finally:
            # Run teardown
            if test.teardown:
                try:
                    if asyncio.iscoroutinefunction(test.teardown):
                        await test.teardown()
                    else:
                        test.teardown()
                except Exception as e:
                    logger.warning(f"Teardown failed for {test.name}: {e}")

    async def run_all(self, report_name: Optional[str] = None) -> TestReport:
        """
        Run all loaded tests.

        Args:
            report_name: Optional report name

        Returns:
            Test report
        """
        return await self.run_tests(
            list(self._tests.keys()),
            report_name=report_name or "All Tests",
        )

    async def run_tests(
        self,
        test_names: List[str],
        report_name: Optional[str] = None,
    ) -> TestReport:
        """
        Run specific tests by name.

        Args:
            test_names: List of test names
            report_name: Optional report name

        Returns:
            Test report
        """
        report = TestReport(
            name=report_name or f"Test Run {uuid.uuid4().hex[:8]}",
            started_at=datetime.now(),
        )

        tests = [self._tests[name] for name in test_names if name in self._tests]

        if self._parallel:
            # Run tests in parallel
            semaphore = asyncio.Semaphore(self._max_parallel)

            async def run_with_semaphore(test: TestCase) -> TestResult:
                async with semaphore:
                    return await self.run_test(test)

            results = await asyncio.gather(
                *[run_with_semaphore(test) for test in tests],
                return_exceptions=True,
            )

            for result in results:
                if isinstance(result, Exception):
                    report.results.append(TestResult(
                        test_name="Unknown",
                        status=TestStatus.ERROR,
                        error=str(result),
                    ))
                else:
                    report.results.append(result)
        else:
            # Run tests sequentially
            for test in tests:
                result = await self.run_test(test)
                report.results.append(result)

        report.finished_at = datetime.now()
        return report

    async def run_by_tags(
        self,
        tags: List[str],
        match_all: bool = False,
        report_name: Optional[str] = None,
    ) -> TestReport:
        """
        Run tests matching tags.

        Args:
            tags: Tags to match
            match_all: If True, test must have all tags
            report_name: Optional report name

        Returns:
            Test report
        """
        matching = []
        for name, test in self._tests.items():
            if match_all:
                if all(tag in test.tags for tag in tags):
                    matching.append(name)
            else:
                if any(tag in test.tags for tag in tags):
                    matching.append(name)

        return await self.run_tests(
            matching,
            report_name=report_name or f"Tests with tags: {', '.join(tags)}",
        )

    def get_test_names(self) -> List[str]:
        """Get all test names"""
        return list(self._tests.keys())

    def get_tests_by_tag(self, tag: str) -> List[str]:
        """Get test names with specific tag"""
        return [
            name for name, test in self._tests.items()
            if tag in test.tags
        ]



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Engine SDK

Stable interface for flyto-cloud and other consumers.

Usage:
    from core.engine.sdk import get_sdk, EngineSDK

    sdk = get_sdk()
    result = await sdk.execute(workflow, inputs)
    catalog = sdk.introspect(workflow, node_id)
"""

from .models import (
    # Enums
    ContextLayer,
    EventType,
    ExecutionStatus,
    IntrospectionMode,
    ResolutionMode,
    VarAvailability,
    VarSource,
    # VarCatalog models
    NodeVarInfo,
    PortVarInfo,
    VarCatalog,
    VarInfo,
    # Event models
    EngineEvent,
    # Execution models
    ExecutionOptions,
    ExecutionResult,
    ExecutionTrace,
    NodeSnapshot,
    # Validation models
    ValidationError,
    ValidationResult,
    ValidationWarning,
    # Expression models
    AutocompleteItem,
    AutocompleteResult,
    ExpressionToken,
    ParsedExpression,
)

from .resolver import (
    ExpressionParser,
    ExpressionSyntaxError,
    VariableNotFoundError,
    VariableResolver,
    create_resolver,
)

from .interface import (
    EngineSDK,
    EngineSDKInterface,
    get_sdk,
)

__all__ = [
    # Enums
    "ContextLayer",
    "EventType",
    "ExecutionStatus",
    "IntrospectionMode",
    "ResolutionMode",
    "VarAvailability",
    "VarSource",
    # VarCatalog models
    "NodeVarInfo",
    "PortVarInfo",
    "VarCatalog",
    "VarInfo",
    # Event models
    "EngineEvent",
    # Execution models
    "ExecutionOptions",
    "ExecutionResult",
    "ExecutionTrace",
    "NodeSnapshot",
    # Validation models
    "ValidationError",
    "ValidationResult",
    "ValidationWarning",
    # Expression models
    "AutocompleteItem",
    "AutocompleteResult",
    "ExpressionToken",
    "ParsedExpression",
    # Resolver
    "ExpressionParser",
    "ExpressionSyntaxError",
    "VariableNotFoundError",
    "VariableResolver",
    "create_resolver",
    # SDK
    "EngineSDK",
    "EngineSDKInterface",
    "get_sdk",
]



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
OAuth Data Models

Configuration and token storage dataclasses.
"""

from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional

from .providers import OAuthProvider, PROVIDER_CONFIGS


@dataclass
class OAuthConfig:
    """OAuth configuration for a provider."""
    provider: OAuthProvider
    client_id: str
    client_secret: str
    redirect_uri: str
    scopes: List[str] = field(default_factory=list)
    authorize_url: Optional[str] = None
    token_url: Optional[str] = None
    revoke_url: Optional[str] = None
    scopes_separator: str = " "
    use_pkce: bool = False
    extra_params: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        """Apply provider defaults."""
        if self.provider != OAuthProvider.CUSTOM:
            defaults = PROVIDER_CONFIGS.get(self.provider, {})
            if not self.authorize_url:
                self.authorize_url = defaults.get("authorize_url")
            if not self.token_url:
                self.token_url = defaults.get("token_url")
            if not self.revoke_url:
                self.revoke_url = defaults.get("revoke_url")
            if self.scopes_separator == " ":
                self.scopes_separator = defaults.get("scopes_separator", " ")


@dataclass
class OAuthToken:
    """OAuth token storage with expiration tracking."""
    access_token: str
    token_type: str = "Bearer"
    expires_at: Optional[datetime] = None
    refresh_token: Optional[str] = None
    scope: Optional[str] = None
    extra: Dict[str, Any] = field(default_factory=dict)

    @property
    def is_expired(self) -> bool:
        """Check if token is expired."""
        if not self.expires_at:
            return False
        return datetime.now(timezone.utc) >= self.expires_at

    @property
    def needs_refresh(self) -> bool:
        """Check if token should be refreshed (5 min buffer)."""
        if not self.expires_at:
            return False
        buffer = timedelta(minutes=5)
        return datetime.now(timezone.utc) >= (self.expires_at - buffer)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for storage."""
        return {
            "access_token": self.access_token,
            "token_type": self.token_type,
            "expires_at": self.expires_at.isoformat() if self.expires_at else None,
            "refresh_token": self.refresh_token,
            "scope": self.scope,
            "extra": self.extra,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "OAuthToken":
        """Create from dictionary."""
        expires_at = data.get("expires_at")
        if expires_at and isinstance(expires_at, str):
            expires_at = datetime.fromisoformat(expires_at)

        return cls(
            access_token=data["access_token"],
            token_type=data.get("token_type", "Bearer"),
            expires_at=expires_at,
            refresh_token=data.get("refresh_token"),
            scope=data.get("scope"),
            extra=data.get("extra", {}),
        )

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "OAuthToken":
        """Create from OAuth token response."""
        expires_at = None
        if "expires_in" in data:
            expires_at = datetime.now(timezone.utc) + timedelta(seconds=int(data["expires_in"]))

        # Common fields
        known_fields = {
            "access_token", "token_type", "expires_in", "refresh_token", "scope"
        }
        extra = {k: v for k, v in data.items() if k not in known_fields}

        return cls(
            access_token=data["access_token"],
            token_type=data.get("token_type", "Bearer"),
            expires_at=expires_at,
            refresh_token=data.get("refresh_token"),
            scope=data.get("scope"),
            extra=extra,
        )



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Convert To String Module
Convert any value to string representation.
"""
from typing import Any, Dict
import json

from ...registry import register_module
from ...schema import compose
from ...schema.builders import field
from ...schema.constants import FieldGroup
from ...errors import ValidationError


@register_module(
    module_id='convert.to_string',
    version='1.0.0',
    category='convert',
    tags=['convert', 'string', 'cast', 'type', 'transform'],
    label='To String',
    label_key='modules.convert.to_string.label',
    description='Convert any value to string',
    description_key='modules.convert.to_string.description',
    icon='Type',
    color='#06B6D4',
    input_types=['any'],
    output_types=['string'],

    can_receive_from=['*'],
    can_connect_to=['string.*', 'data.*', 'flow.*'],

    retryable=False,
    concurrent_safe=True,

    requires_credentials=False,
    handles_sensitive_data=False,
    required_permissions=[],

    params_schema=compose(
        field(
            'value',
            type='any',
            label='Value',
            label_key='modules.convert.to_string.params.value.label',
            description='Value to convert',
            description_key='modules.convert.to_string.params.value.description',
            required=True,
            group=FieldGroup.BASIC,
        ),
        field(
            'pretty',
            type='boolean',
            label='Pretty Print',
            label_key='modules.convert.to_string.params.pretty.label',
            description='Format objects/arrays with indentation',
            description_key='modules.convert.to_string.params.pretty.description',
            default=False,
            group=FieldGroup.OPTIONS,
        ),
    ),
    output_schema={
        'result': {
            'type': 'string',
            'description': 'String representation',
            'description_key': 'modules.convert.to_string.output.result.description'
        },
        'original_type': {
            'type': 'string',
            'description': 'Original value type',
            'description_key': 'modules.convert.to_string.output.original_type.description'
        }
    },
    timeout_ms=5000,
)
async def convert_to_string(context: Dict[str, Any]) -> Dict[str, Any]:
    """Convert any value to string."""
    params = context['params']
    value = params.get('value')
    pretty = params.get('pretty', False)

    if value is None:
        result = ''
        original_type = 'null'
    elif isinstance(value, bool):
        result = 'true' if value else 'false'
        original_type = 'boolean'
    elif isinstance(value, (int, float)):
        result = str(value)
        original_type = 'number'
    elif isinstance(value, str):
        result = value
        original_type = 'string'
    elif isinstance(value, (list, dict)):
        if pretty:
            result = json.dumps(value, indent=2, ensure_ascii=False)
        else:
            result = json.dumps(value, ensure_ascii=False)
        original_type = 'array' if isinstance(value, list) else 'object'
    else:
        result = str(value)
        original_type = type(value).__name__

    return {
        'ok': True,
        'data': {
            'result': result,
            'original_type': original_type
        }
    }



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Evidence Store

Stores execution evidence to filesystem.
"""

import json
import logging
from pathlib import Path
from typing import Any, List, Optional

from .models import StepEvidence

logger = logging.getLogger(__name__)


class EvidenceStore:
    """
    Stores execution evidence to filesystem.

    Directory structure:
        evidence/
        ├── exec_abc123/
        │   ├── evidence.jsonl    # All step metadata
        │   ├── step_1.png        # Screenshot
        │   ├── step_1.html       # DOM snapshot
        │   └── ...
        └── exec_def456/
            └── ...
    """

    def __init__(
        self,
        base_path: Path,
        capture_context: bool = True,
        max_context_depth: int = 5,
    ):
        """
        Initialize evidence store.

        Args:
            base_path: Base directory for evidence storage
            capture_context: Whether to capture context snapshots
            max_context_depth: Max nesting depth for context serialization
        """
        self.base_path = Path(base_path)
        self.capture_context = capture_context
        self.max_context_depth = max_context_depth

    def get_execution_dir(self, execution_id: str) -> Path:
        """Get/create directory for execution evidence"""
        path = self.base_path / execution_id
        path.mkdir(parents=True, exist_ok=True)
        return path

    async def save_evidence(self, evidence: StepEvidence) -> None:
        """
        Save evidence metadata to JSONL file.

        Appends to evidence.jsonl in the execution directory.
        """
        try:
            exec_dir = self.get_execution_dir(evidence.execution_id)
            jsonl_path = exec_dir / "evidence.jsonl"

            # Serialize to JSON
            data = evidence.to_dict()
            # Truncate large context if needed
            data = self._truncate_large_values(data)
            line = json.dumps(data, default=str, ensure_ascii=False) + '\n'

            # Append to JSONL file
            with open(jsonl_path, 'a', encoding='utf-8') as f:
                f.write(line)

            logger.debug(f"Saved evidence for step {evidence.step_id}")

        except Exception as e:
            logger.warning(f"Failed to save evidence for {evidence.step_id}: {e}")

    async def save_screenshot(
        self,
        execution_id: str,
        step_id: str,
        screenshot_bytes: bytes,
    ) -> str:
        """
        Save screenshot and return relative path.

        Args:
            execution_id: Execution ID
            step_id: Step ID (used as filename)
            screenshot_bytes: PNG image data

        Returns:
            Relative filename (e.g., "step_1.png")
        """
        try:
            exec_dir = self.get_execution_dir(execution_id)
            filename = f"{step_id}.png"
            path = exec_dir / filename

            with open(path, 'wb') as f:
                f.write(screenshot_bytes)

            logger.debug(f"Saved screenshot: {path}")
            return filename

        except Exception as e:
            logger.warning(f"Failed to save screenshot for {step_id}: {e}")
            return ""

    async def save_dom_snapshot(
        self,
        execution_id: str,
        step_id: str,
        dom_html: str,
    ) -> str:
        """
        Save DOM snapshot and return relative path.

        Args:
            execution_id: Execution ID
            step_id: Step ID (used as filename)
            dom_html: HTML content

        Returns:
            Relative filename (e.g., "step_1.html")
        """
        try:
            exec_dir = self.get_execution_dir(execution_id)
            filename = f"{step_id}.html"
            path = exec_dir / filename

            with open(path, 'w', encoding='utf-8') as f:
                f.write(dom_html)

            logger.debug(f"Saved DOM snapshot: {path}")
            return filename

        except Exception as e:
            logger.warning(f"Failed to save DOM snapshot for {step_id}: {e}")
            return ""

    async def load_evidence(self, execution_id: str) -> List[StepEvidence]:
        """
        Load all evidence for an execution.

        Args:
            execution_id: Execution ID to load

        Returns:
            List of StepEvidence records in execution order
        """
        exec_dir = self.base_path / execution_id
        jsonl_path = exec_dir / "evidence.jsonl"

        if not jsonl_path.exists():
            return []

        evidence_list = []
        try:
            with open(jsonl_path, 'r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if line:
                        data = json.loads(line)
                        evidence_list.append(StepEvidence.from_dict(data))
        except Exception as e:
            logger.warning(f"Failed to load evidence for {execution_id}: {e}")

        return evidence_list

    async def load_step_evidence(
        self,
        execution_id: str,
        step_id: str,
    ) -> Optional[StepEvidence]:
        """
        Load evidence for a specific step.

        Args:
            execution_id: Execution ID
            step_id: Step ID to find

        Returns:
            StepEvidence if found, None otherwise
        """
        all_evidence = await self.load_evidence(execution_id)
        for evidence in all_evidence:
            if evidence.step_id == step_id:
                return evidence
        return None

    async def get_screenshot_path(
        self,
        execution_id: str,
        step_id: str,
    ) -> Optional[Path]:
        """Get full path to screenshot file if it exists"""
        path = self.base_path / execution_id / f"{step_id}.png"
        return path if path.exists() else None

    async def get_dom_snapshot_path(
        self,
        execution_id: str,
        step_id: str,
    ) -> Optional[Path]:
        """Get full path to DOM snapshot file if it exists"""
        path = self.base_path / execution_id / f"{step_id}.html"
        return path if path.exists() else None

    async def list_executions(self) -> List[str]:
        """List all execution IDs with evidence"""
        if not self.base_path.exists():
            return []

        return [
            d.name for d in self.base_path.iterdir()
            if d.is_dir() and (d / "evidence.jsonl").exists()
        ]

    async def delete_execution(self, execution_id: str) -> bool:
        """
        Delete all evidence for an execution.

        Args:
            execution_id: Execution ID to delete

        Returns:
            True if deleted, False if not found
        """
        import shutil
        exec_dir = self.base_path / execution_id
        if exec_dir.exists():
            shutil.rmtree(exec_dir)
            return True
        return False

    def _truncate_large_values(
        self,
        data: Any,
        max_str_length: int = 10000,
        current_depth: int = 0,
    ) -> Any:
        """Truncate large string values to prevent huge JSONL files"""
        if current_depth > self.max_context_depth:
            return "[truncated: max depth]"

        if isinstance(data, dict):
            return {
                k: self._truncate_large_values(v, max_str_length, current_depth + 1)
                for k, v in data.items()
            }
        elif isinstance(data, list):
            if len(data) > 100:
                return data[:100] + ["[truncated: list too long]"]
            return [
                self._truncate_large_values(v, max_str_length, current_depth + 1)
                for v in data
            ]
        elif isinstance(data, str) and len(data) > max_str_length:
            return data[:max_str_length] + f"... [truncated: {len(data)} chars]"
        elif isinstance(data, bytes):
            return f"[binary data: {len(data)} bytes]"
        return data



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Docker PS Module
List Docker containers
"""
import asyncio
import json
import logging
from typing import Any, Dict, List

from ...registry import register_module
from ...schema import compose
from ...schema.builders import field
from ...schema.constants import FieldGroup
from ...errors import ModuleError

logger = logging.getLogger(__name__)


def _parse_container_line(line: str) -> Dict[str, Any]:
    """Parse a single JSON line from docker ps --format json."""
    try:
        raw = json.loads(line)
    except (json.JSONDecodeError, TypeError):
        return {}

    return {
        'id': raw.get('ID', ''),
        'name': raw.get('Names', ''),
        'image': raw.get('Image', ''),
        'status': raw.get('Status', ''),
        'ports': raw.get('Ports', ''),
        'state': raw.get('State', ''),
        'created': raw.get('CreatedAt', ''),
    }


@register_module(
    module_id='docker.ps',
    version='1.0.0',
    category='docker',
    tags=['docker', 'container', 'list', 'ps', 'devops'],
    label='List Docker Containers',
    label_key='modules.docker.ps.label',
    description='List Docker containers',
    description_key='modules.docker.ps.description',
    icon='Container',
    color='#0DB7ED',
    input_types=[],
    output_types=['array'],

    can_receive_from=['*'],
    can_connect_to=['*'],

    retryable=True,
    concurrent_safe=True,

    requires_credentials=False,
    handles_sensitive_data=False,
    required_permissions=['docker.read'],

    params_schema=compose(
        field(
            'all',
            type='boolean',
            label='Show All',
            label_key='modules.docker.ps.params.all.label',
            description='Show all containers (default shows just running)',
            description_key='modules.docker.ps.params.all.description',
            default=False,
            group=FieldGroup.BASIC,
        ),
        field(
            'filters',
            type='object',
            label='Filters',
            label_key='modules.docker.ps.params.filters.label',
            description='Filter containers (e.g. {"name": "my-app", "status": "running"})',
            description_key='modules.docker.ps.params.filters.description',
            group=FieldGroup.OPTIONS,
        ),
    ),
    output_schema={
        'containers': {
            'type': 'array',
            'description': 'List of containers with id, name, image, status, ports',
            'description_key': 'modules.docker.ps.output.containers.description',
        },
        'count': {
            'type': 'number',
            'description': 'Number of containers found',
            'description_key': 'modules.docker.ps.output.count.description',
        },
    },
    examples=[
        {
            'title': 'List running containers',
            'params': {},
        },
        {
            'title': 'List all containers',
            'params': {'all': True},
        },
        {
            'title': 'Filter by name',
            'params': {'filters': {'name': 'nginx'}},
        },
    ],
    timeout_ms=30000,
)
async def docker_ps(context: Dict[str, Any]) -> Dict[str, Any]:
    """List Docker containers."""
    params = context.get('params', {})

    show_all = params.get('all', False)
    filters = params.get('filters') or {}

    args = ['docker', 'ps', '--format', '{{json .}}', '--no-trunc']

    if show_all:
        args.append('--all')

    if isinstance(filters, dict):
        for key, value in filters.items():
            args.extend(['--filter', '%s=%s' % (str(key), str(value))])

    logger.info("Docker ps: %s", ' '.join(args))

    try:
        process = await asyncio.create_subprocess_exec(
            *args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        try:
            stdout_bytes, stderr_bytes = await asyncio.wait_for(
                process.communicate(),
                timeout=25,
            )
        except asyncio.TimeoutError:
            process.kill()
            await process.wait()
            raise ModuleError("Docker ps timed out")

        stdout = stdout_bytes.decode('utf-8', errors='replace').strip()
        stderr = stderr_bytes.decode('utf-8', errors='replace').strip()

        if process.returncode != 0:
            error_msg = stderr if stderr else stdout
            raise ModuleError(
                "Docker ps failed (exit code %d): %s" % (process.returncode, error_msg)
            )

        containers: List[Dict[str, Any]] = []
        if stdout:
            for line in stdout.splitlines():
                line = line.strip()
                if not line:
                    continue
                parsed = _parse_container_line(line)
                if parsed:
                    containers.append(parsed)

        return {
            'ok': True,
            'data': {
                'containers': containers,
                'count': len(containers),
            },
        }

    except ModuleError:
        raise
    except Exception as e:
        logger.error("Docker ps error: %s", e)
        raise ModuleError("Docker ps failed: %s" % str(e))



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Utility Modules
Helper modules for delays, random data, date/time operations, etc.
"""
from typing import Any, Dict
from ...base import BaseModule
from ...registry import register_module
import asyncio
import random
import string
from datetime import datetime, timedelta
import hashlib
import uuid


@register_module(
    module_id='utility.datetime.now',
    version='1.0.0',
    category='utility',
    tags=['utility', 'datetime', 'time', 'date', 'timestamp'],
    label='Current Date/Time',
    label_key='modules.utility.datetime.now.label',
    description='Get current date and time',
    description_key='modules.utility.datetime.now.description',
    icon='Calendar',
    color='#3B82F6',

    # Connection types
    input_types=[],
    output_types=['string'],


    can_receive_from=['*'],
    can_connect_to=['data.*', 'string.*', 'file.*', 'http.*', 'notify.*', 'flow.*', 'utility.*'],    # Phase 2: Execution settings
    retryable=False,  # Time operations are instant and deterministic
    concurrent_safe=True,  # Multiple time operations can run in parallel

    # Phase 2: Security settings
    requires_credentials=False,  # No credentials needed
    handles_sensitive_data=False,  # No sensitive data handled
    required_permissions=[],  # No special permissions needed

    params_schema={
        'format': {
            'type': 'select',
            'label': 'Format',
            'label_key': 'modules.utility.datetime.now.params.format.label',
            'description': 'Output format',
            'description_key': 'modules.utility.datetime.now.params.format.description',
            'options': [
                {'value': 'iso', 'label': 'ISO 8601 (2024-01-15T10:30:00)'},
                {'value': 'unix', 'label': 'Unix timestamp (seconds)'},
                {'value': 'unix_ms', 'label': 'Unix timestamp (milliseconds)'},
                {'value': 'date', 'label': 'Date only (2024-01-15)'},
                {'value': 'time', 'label': 'Time only (10:30:00)'},
                {'value': 'custom', 'label': 'Custom format'}
            ],
            'default': 'iso',
            'required': False
        },
        'custom_format': {
            'type': 'string',
            'label': 'Custom Format',
            'label_key': 'modules.utility.datetime.now.params.custom_format.label',
            'description': 'Python strftime format (if format=custom)',
            'description_key': 'modules.utility.datetime.now.params.custom_format.description',
            'placeholder': '%Y-%m-%d %H:%M:%S',
            'required': False
        },
        'timezone': {
            'type': 'string',
            'label': 'Timezone',
            'label_key': 'modules.utility.datetime.now.params.timezone.label',
            'description': 'Timezone (default: UTC)',
            'description_key': 'modules.utility.datetime.now.params.timezone.description',
            'placeholder': 'UTC',
            'default': 'UTC',
            'required': False
        }
    },
    output_schema={
        'status': {'type': 'string', 'description': 'Operation status (success/error)',
                'description_key': 'modules.utility.datetime.now.output.status.description'},
        'datetime': {'type': 'string', 'description': 'Formatted date/time',
                'description_key': 'modules.utility.datetime.now.output.datetime.description'},
        'timestamp': {'type': 'number', 'description': 'Unix timestamp',
                'description_key': 'modules.utility.datetime.now.output.timestamp.description'},
        'iso': {'type': 'string', 'description': 'ISO format',
                'description_key': 'modules.utility.datetime.now.output.iso.description'}
    },
    examples=[
        {
            'name': 'Get current ISO datetime',
            'params': {
                'format': 'iso'
            }
        },
        {
            'name': 'Get Unix timestamp',
            'params': {
                'format': 'unix'
            }
        }
    ],
    author='Flyto Team',
    license='MIT',
    timeout_ms=5000,
)
class DateTimeNowModule(BaseModule):
    """Get current date/time"""

    module_name = "Current Date/Time"
    module_description = "Get current date and time in various formats"

    def validate_params(self) -> None:
        self.format = self.params.get('format', 'iso')
        self.custom_format = self.params.get('custom_format', '%Y-%m-%d %H:%M:%S')
        self.timezone = self.params.get('timezone', 'UTC')

    async def execute(self) -> Any:
        now = datetime.utcnow()  # Always use UTC for consistency

        # Format output
        if self.format == 'iso':
            formatted = now.isoformat()
        elif self.format == 'unix':
            formatted = int(now.timestamp())
        elif self.format == 'unix_ms':
            formatted = int(now.timestamp() * 1000)
        elif self.format == 'date':
            formatted = now.strftime('%Y-%m-%d')
        elif self.format == 'time':
            formatted = now.strftime('%H:%M:%S')
        elif self.format == 'custom':
            formatted = now.strftime(self.custom_format)
        else:
            formatted = now.isoformat()

        return {
            'status': 'success',
            'datetime': formatted,
            'timestamp': int(now.timestamp()),
            'iso': now.isoformat()
        }





# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Template Modules

Provides modules for executing templates as nodes in workflows.
Allows templates from user's library to be used as reusable components.
"""

from .invoke import InvokeTemplate

__all__ = ['InvokeTemplate']



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Datetime Operations Modules

Provides date and time manipulation capabilities.
"""
from typing import Any, Dict
from ...base import BaseModule
from ...registry import register_module
from ...schema import compose, presets
from datetime import datetime, timedelta
import time


@register_module(
    module_id='datetime.add',
    version='1.0.0',
    category='utility',
    subcategory='datetime',
    tags=['datetime', 'add', 'date', 'time'],
    label='Add Time',
    label_key='modules.datetime.add.label',
    description='Add time to datetime',
    description_key='modules.datetime.add.description',
    icon='Plus',
    color='#8B5CF6',

    # Connection types
    input_types=['datetime', 'string'],
    output_types=['datetime', 'string'],


    can_receive_from=['*'],
    can_connect_to=['data.*', 'string.*', 'file.*', 'http.*', 'notify.*', 'flow.*', 'utility.*'],    # Phase 2: Execution settings
    timeout_ms=5000,
    retryable=False,
    concurrent_safe=True,

    # Phase 2: Security settings
    requires_credentials=False,
    handles_sensitive_data=False,
    required_permissions=[],

    # Schema-driven params
    params_schema=compose(
        presets.DATETIME_INPUT(default='now'),
        presets.TIME_DAYS(default=0),
        presets.TIME_HOURS(default=0),
        presets.TIME_MINUTES(default=0),
        presets.TIME_SECONDS(default=0),
    ),
    output_schema={
        'result': {'type': 'string', 'description': 'The operation result',
                'description_key': 'modules.datetime.add.output.result.description'},
        'timestamp': {'type': 'number', 'description': 'Unix timestamp',
                'description_key': 'modules.datetime.add.output.timestamp.description'}
    },
    examples=[
        {
            'title': 'Add 7 days',
            'params': {
                'datetime': 'now',
                'days': 7
            }
        },
        {
            'title': 'Add 2 hours 30 minutes',
            'params': {
                'datetime': '2024-01-15T10:00:00',
                'hours': 2,
                'minutes': 30
            }
        }
    ],
    author='Flyto Team',
    license='MIT'
)
class DateTimeAddModule(BaseModule):
    """DateTime Add Module"""

    def validate_params(self) -> None:
        self.datetime_str = self.params.get('datetime', 'now')
        self.days = self.params.get('days', 0)
        self.hours = self.params.get('hours', 0)
        self.minutes = self.params.get('minutes', 0)
        self.seconds = self.params.get('seconds', 0)

    async def execute(self) -> Any:
        # Parse datetime
        if self.datetime_str == 'now':
            dt = datetime.now()
        else:
            try:
                dt = datetime.fromisoformat(self.datetime_str.replace('Z', '+00:00'))
            except (ValueError, AttributeError) as e:
                raise ValueError(
                    f"Invalid datetime format: '{self.datetime_str}'. "
                    f"Expected ISO 8601 format (e.g., '2024-01-15T10:30:00Z'). "
                    f"Error: {e}"
                )

        # Add time
        delta = timedelta(
            days=self.days,
            hours=self.hours,
            minutes=self.minutes,
            seconds=self.seconds
        )
        result_dt = dt + delta

        return {
            "result": result_dt.isoformat(),
            "timestamp": result_dt.timestamp()
        }





# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Database Connectors

PostgreSQL, MySQL, and MongoDB integrations.
"""

from .postgresql import postgresql_query
from .mysql import mysql_query
from .mongodb_find import mongodb_find
from .mongodb_insert import mongodb_insert

__all__ = [
    'postgresql_query',
    'mysql_query',
    'mongodb_find',
    'mongodb_insert',
]



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Severity Policy

Defines severity policies for different CI/CD gate levels.
"""
from dataclasses import dataclass, field
from typing import Dict, Optional, Set

from .types import GateLevel, Severity


@dataclass
class SeverityPolicy:
    """
    Severity policy for a specific gate level.

    Determines which severity levels block validation based on the gate.
    """
    gate: GateLevel
    blocking_severities: Set[Severity] = field(default_factory=set)
    upgrade_severities: Dict[Severity, Severity] = field(default_factory=dict)
    stable_only_strict: bool = False

    def should_block(self, severity: Severity, stability: str = "stable") -> bool:
        """
        Check if an issue should block based on severity and module stability.

        Args:
            severity: The issue severity
            stability: The module stability level (stable, beta, alpha, etc.)

        Returns:
            True if the issue should block validation
        """
        # Apply severity upgrade if configured
        effective_severity = self.upgrade_severities.get(severity, severity)

        # For stable_only_strict, non-stable modules are less strict
        if self.stable_only_strict and stability != "stable":
            # Only BLOCKER and FATAL block for non-stable modules
            return effective_severity in (Severity.BLOCKER, Severity.FATAL)

        return effective_severity in self.blocking_severities

    def get_effective_severity(self, severity: Severity) -> Severity:
        """Get the effective severity after applying upgrades."""
        return self.upgrade_severities.get(severity, severity)


# Predefined policies for each gate level

DEV_POLICY = SeverityPolicy(
    gate=GateLevel.DEV,
    blocking_severities={Severity.FATAL},
    upgrade_severities={},
    stable_only_strict=False,
)

CI_POLICY = SeverityPolicy(
    gate=GateLevel.CI,
    blocking_severities={Severity.ERROR, Severity.BLOCKER, Severity.FATAL},
    upgrade_severities={},
    stable_only_strict=True,
)

RELEASE_POLICY = SeverityPolicy(
    gate=GateLevel.RELEASE,
    blocking_severities={Severity.ERROR, Severity.BLOCKER, Severity.FATAL},
    upgrade_severities={},
    stable_only_strict=False,
)

STRICT_POLICY = SeverityPolicy(
    gate=GateLevel.STRICT,
    blocking_severities={Severity.INFO, Severity.WARN, Severity.ERROR, Severity.BLOCKER, Severity.FATAL},
    upgrade_severities={
        Severity.INFO: Severity.WARN,
        Severity.WARN: Severity.ERROR,
    },
    stable_only_strict=False,
)


def get_policy(gate: GateLevel) -> SeverityPolicy:
    """Get the predefined policy for a gate level."""
    policies = {
        GateLevel.DEV: DEV_POLICY,
        GateLevel.CI: CI_POLICY,
        GateLevel.RELEASE: RELEASE_POLICY,
        GateLevel.STRICT: STRICT_POLICY,
    }
    return policies.get(gate, CI_POLICY)



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Unit Tests for Introspection

Tests VarCatalog building for workflow variable introspection.
"""

import pytest
from typing import Any, Dict, List

from core.engine.sdk.models import (
    IntrospectionMode,
    VarAvailability,
    VarSource,
    VarCatalog,
    VarInfo,
    PortVarInfo,
    NodeVarInfo,
)
from core.engine.introspection.catalog import (
    build_catalog,
)


# =============================================================================
# Test Fixtures
# =============================================================================

def create_simple_workflow() -> Dict[str, Any]:
    """Create a simple linear workflow for testing"""
    return {
        "id": "test-workflow",
        "nodes": [
            {
                "id": "trigger",
                "module_id": "core.trigger.webhook",
                "params": {},
            },
            {
                "id": "process",
                "module_id": "data.transform.map",
                "params": {"expression": "{{input}}"},
            },
            {
                "id": "output",
                "module_id": "core.output.json",
                "params": {},
            },
        ],
        "edges": [
            {"source": "trigger", "target": "process"},
            {"source": "process", "target": "output"},
        ],
    }


def create_branching_workflow() -> Dict[str, Any]:
    """Create a workflow with branches"""
    return {
        "id": "branching-workflow",
        "nodes": [
            {
                "id": "start",
                "module_id": "core.trigger.manual",
                "params": {},
            },
            {
                "id": "condition",
                "module_id": "flow.condition",
                "params": {"condition": "{{input.status}} == 'active'"},
            },
            {
                "id": "branch_a",
                "module_id": "data.transform.map",
                "params": {},
            },
            {
                "id": "branch_b",
                "module_id": "data.transform.map",
                "params": {},
            },
            {
                "id": "merge",
                "module_id": "flow.merge",
                "params": {},
            },
        ],
        "edges": [
            {"source": "start", "target": "condition"},
            {"source": "condition", "target": "branch_a", "sourceHandle": "true"},
            {"source": "condition", "target": "branch_b", "sourceHandle": "false"},
            {"source": "branch_a", "target": "merge"},
            {"source": "branch_b", "target": "merge"},
        ],
    }


# =============================================================================
# CatalogBuilder Tests
# =============================================================================

class TestCatalogBuilder:
    """Tests for build_catalog function"""

    def test_build_catalog_edit_mode(self):
        workflow = create_simple_workflow()
        catalog = build_catalog(
            workflow=workflow,
            node_id="process",
            mode=IntrospectionMode.EDIT,
        )

        assert isinstance(catalog, VarCatalog)
        assert catalog.mode == IntrospectionMode.EDIT

    def test_build_catalog_runtime_mode(self):
        workflow = create_simple_workflow()
        context = {
            "trigger": {"ok": True, "data": {"message": "hello"}},
        }
        catalog = build_catalog(
            workflow=workflow,
            node_id="process",
            mode=IntrospectionMode.RUNTIME,
            context_snapshot=context,
        )

        assert catalog.mode == IntrospectionMode.RUNTIME

    def test_catalog_contains_upstream_nodes(self):
        workflow = create_simple_workflow()
        catalog = build_catalog(
            workflow=workflow,
            node_id="output",
            mode=IntrospectionMode.EDIT,
        )

        # Should have outputs from trigger and process
        assert "trigger" in catalog.nodes
        assert "process" in catalog.nodes

    def test_catalog_excludes_downstream_nodes(self):
        workflow = create_simple_workflow()
        catalog = build_catalog(
            workflow=workflow,
            node_id="trigger",
            mode=IntrospectionMode.EDIT,
        )

        # Should NOT have process or output (they're downstream)
        assert "process" not in catalog.nodes
        assert "output" not in catalog.nodes


# =============================================================================
# VarCatalog Model Tests
# =============================================================================

class TestVarCatalogModel:
    """Tests for VarCatalog data model"""

    def test_var_info_creation(self):
        var = VarInfo(
            path="input.data",
            var_type="object",
        )

        assert var.path == "input.data"
        assert var.var_type == "object"

    def test_port_var_info_creation(self):
        port = PortVarInfo(
            port_id="output",
            var_type="object",
            description="Output port",
        )

        assert port.port_id == "output"
        assert port.var_type == "object"

    def test_node_var_info_creation(self):
        node = NodeVarInfo(
            node_id="fetch",
            node_type="http.request",
        )

        assert node.node_id == "fetch"
        assert node.node_type == "http.request"

    def test_var_catalog_creation(self):
        catalog = VarCatalog(
            node_id="process",
            mode=IntrospectionMode.EDIT,
        )

        assert catalog.node_id == "process"
        assert catalog.mode == IntrospectionMode.EDIT


# =============================================================================
# Build Catalog Integration Tests
# =============================================================================

class TestBuildCatalogIntegration:
    """Integration tests for build_catalog function"""

    def test_build_catalog_steps_format(self):
        """Test catalog building from steps format (alternative to nodes/edges)"""
        workflow = {
            "id": "test",
            "steps": [
                {
                    "id": "step1",
                    "module_id": "http.request",
                    "params": {"url": "https://example.com"},
                },
                {
                    "id": "step2",
                    "module_id": "data.transform.map",
                    "params": {"expression": "{{step1.data}}"},
                },
            ],
        }

        catalog = build_catalog(
            workflow=workflow,
            node_id="step2",
            mode=IntrospectionMode.EDIT,
        )

        # Should have step1 as upstream
        assert "step1" in catalog.nodes

    def test_build_catalog_empty_workflow(self):
        """Test with empty workflow"""
        workflow = {"id": "empty", "nodes": [], "edges": []}

        # Should not crash
        catalog = build_catalog(
            workflow=workflow,
            node_id="nonexistent",
            mode=IntrospectionMode.EDIT,
        )
        assert isinstance(catalog, VarCatalog)



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Statistics Mode Module
Calculate mode (most frequent value) of data.
"""
from typing import Any, Dict, List
from collections import Counter

from ...registry import register_module
from ...schema import compose
from ...schema.builders import field
from ...schema.constants import FieldGroup
from ...errors import ValidationError


@register_module(
    module_id='stats.mode',
    version='1.0.0',
    category='stats',
    tags=['stats', 'mode', 'frequency', 'math', 'advanced'],
    label='Mode',
    label_key='modules.stats.mode.label',
    description='Calculate mode (most frequent value)',
    description_key='modules.stats.mode.description',
    icon='Calculator',
    color='#3B82F6',
    input_types=['array'],
    output_types=['any'],

    can_receive_from=['*'],
    can_connect_to=['math.*', 'data.*', 'flow.*', 'string.*', 'notify.*'],

    retryable=False,
    concurrent_safe=True,

    requires_credentials=False,
    handles_sensitive_data=False,
    required_permissions=[],

    params_schema=compose(
        field(
            'values',
            type='array',
            label='Values',
            label_key='modules.stats.mode.params.values.label',
            description='Array of values',
            description_key='modules.stats.mode.params.values.description',
            required=True,
            placeholder='[1, 2, 2, 3, 3, 3]',
            group=FieldGroup.BASIC,
        ),
        field(
            'all_modes',
            type='boolean',
            label='Return All Modes',
            label_key='modules.stats.mode.params.all_modes.label',
            description='Return all modes if multiple exist',
            description_key='modules.stats.mode.params.all_modes.description',
            default=False,
            group=FieldGroup.OPTIONS,
        ),
    ),
    output_schema={
        'mode': {
            'type': 'any',
            'description': 'Most frequent value(s)',
            'description_key': 'modules.stats.mode.output.mode.description'
        },
        'frequency': {
            'type': 'number',
            'description': 'Frequency of mode',
            'description_key': 'modules.stats.mode.output.frequency.description'
        },
        'count': {
            'type': 'number',
            'description': 'Number of values',
            'description_key': 'modules.stats.mode.output.count.description'
        }
    },
    timeout_ms=5000,
)
async def stats_mode(context: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate mode of values."""
    params = context['params']
    values = params.get('values')
    all_modes = params.get('all_modes', False)

    if values is None:
        raise ValidationError("Missing required parameter: values", field="values")

    if not isinstance(values, list):
        raise ValidationError("Parameter must be an array", field="values")

    if len(values) == 0:
        raise ValidationError("Array cannot be empty", field="values")

    # Convert unhashable types to strings for counting
    def make_hashable(v):
        if isinstance(v, (list, dict)):
            return str(v)
        return v

    hashable_values = [make_hashable(v) for v in values]
    counter = Counter(hashable_values)
    max_freq = max(counter.values())

    modes = [v for v, freq in counter.items() if freq == max_freq]

    if all_modes:
        mode = modes
    else:
        mode = modes[0]

    return {
        'ok': True,
        'data': {
            'mode': mode,
            'frequency': max_freq,
            'count': len(values)
        }
    }



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Retry Logic

Step execution retry with backoff strategies.
"""

import asyncio
import logging
from typing import Any, Callable, Coroutine, Dict, Optional

from ..exceptions import StepExecutionError, StepTimeoutError
from ..hooks import ExecutorHooks, HookAction
from ...constants import (
    DEFAULT_MAX_RETRIES,
    DEFAULT_RETRY_DELAY_MS,
    EXPONENTIAL_BACKOFF_BASE,
)
from .context_builder import create_step_context

logger = logging.getLogger(__name__)


async def execute_with_retry(
    step_id: str,
    execute_fn: Callable[[], Coroutine[Any, Any, Any]],
    retry_config: Dict[str, Any],
    hooks: Optional[ExecutorHooks] = None,
    step_config: Optional[Dict[str, Any]] = None,
    step_index: int = 0,
    context: Optional[Dict[str, Any]] = None,
    workflow_id: str = "unknown",
    workflow_name: str = "Unnamed Workflow",
    total_steps: int = 0,
) -> Any:
    """
    Execute with retry logic and optional timeout per attempt.

    Args:
        step_id: ID of the step
        execute_fn: Async function to execute
        retry_config: Retry configuration (count, delay_ms, backoff)
        hooks: Optional executor hooks
        step_config: Full step configuration (for hooks)
        step_index: Index of the step
        context: Current workflow context
        workflow_id: Parent workflow ID
        workflow_name: Parent workflow name
        total_steps: Total steps in workflow

    Returns:
        Execution result

    Raises:
        StepExecutionError: If all retry attempts fail
    """
    from ..hooks import NullHooks
    hooks = hooks or NullHooks()
    context = context or {}

    max_retries = retry_config.get('count', DEFAULT_MAX_RETRIES)
    delay_ms = retry_config.get('delay_ms', DEFAULT_RETRY_DELAY_MS)
    backoff = retry_config.get('backoff', 'linear')

    last_error: Optional[Exception] = None

    for attempt in range(max_retries + 1):
        try:
            return await execute_fn()
        except (StepTimeoutError, StepExecutionError, Exception) as e:
            last_error = e

            if attempt < max_retries:
                wait_time = _calculate_wait_time(delay_ms, attempt, backoff)

                logger.warning(
                    f"Step '{step_id}' failed (attempt {attempt + 1}/{max_retries + 1}). "
                    f"Retrying in {wait_time:.1f}s..."
                )

                # Call retry hook
                if step_config:
                    retry_context = create_step_context(
                        workflow_id=workflow_id,
                        workflow_name=workflow_name,
                        total_steps=total_steps,
                        step_config=step_config,
                        step_index=step_index,
                        context=context,
                        error=e,
                        attempt=attempt + 2,
                        max_attempts=max_retries + 1,
                    )
                    retry_result = hooks.on_retry(retry_context)
                    if retry_result.action == HookAction.ABORT:
                        raise StepExecutionError(
                            step_id,
                            f"Retry aborted by hook: {retry_result.abort_reason}",
                            e
                        )

                await asyncio.sleep(wait_time)
            else:
                logger.error(f"Step '{step_id}' failed after {max_retries + 1} attempts")

    raise StepExecutionError(
        step_id,
        f"Step failed after {max_retries + 1} attempts",
        last_error
    )


def _calculate_wait_time(delay_ms: int, attempt: int, backoff: str) -> float:
    """Calculate wait time based on backoff strategy."""
    if backoff == 'exponential':
        return (delay_ms / 1000) * (EXPONENTIAL_BACKOFF_BASE ** attempt)
    elif backoff == 'linear':
        return (delay_ms / 1000) * (attempt + 1)
    else:
        return delay_ms / 1000



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Connection Rules Definitions

Default connection rules by category.
These are STRICT defaults - individual modules can override with more permissive rules if needed.
"""

from typing import Dict

from .models import ConnectionRule


# Special nodes
SPECIAL_NODES = {"start", "end", "trigger"}


# Default rules by category
# Pattern format: "category.*" or "category.specific" or "*" for any

CONNECTION_RULES: Dict[str, ConnectionRule] = {
    # =========================================================================
    # Browser Automation - STRICT: Must stay in browser context chain
    # =========================================================================
    "browser": ConnectionRule(
        category="browser",
        can_connect_to=[
            "browser.*",    # Chain browser actions
            "element.*",    # Element operations
            "flow.*",       # Flow control
        ],
        can_receive_from=[
            "browser.*",    # Chain browser actions
            "flow.*",       # From flow control
            "start",        # Can be first node
        ],
        description="Browser modules must stay in browser context - NO direct connection to AI/API"
    ),

    # =========================================================================
    # Element Operations - STRICT: Requires browser context
    # =========================================================================
    "element": ConnectionRule(
        category="element",
        can_connect_to=[
            "element.*",    # Chain element actions
            "browser.*",    # Back to browser
            "data.*",       # Extract data
            "string.*",     # String operations
            "flow.*",       # Flow control
            "file.*",       # Save data
        ],
        can_receive_from=[
            "browser.*",    # From browser (find, goto, etc.)
            "element.*",    # Chain elements
            "flow.*",       # From flow control
        ],
        description="Element modules require browser context - can only receive from browser/element"
    ),

    # =========================================================================
    # Flow Control - RESTRICTED: Different rules for different flow modules
    # Category default is restrictive; specific modules override in their definition
    # =========================================================================
    "flow": ConnectionRule(
        category="flow",
        can_connect_to=["*"],  # Flow outputs can go anywhere
        can_receive_from=[
            # Flow control should receive from data-producing modules
            "data.*",
            "http.*",
            "string.*",
            "array.*",
            "object.*",
            "math.*",
            "file.*",
            "database.*",
            "ai.*",
            "flow.*",       # Chain flow controls
            "element.*",    # Element data
            "start",
        ],
        description="Flow control receives from data-producing modules"
    ),

    # =========================================================================
    # Data Transformation - Universal but NO browser output
    # =========================================================================
    "data": ConnectionRule(
        category="data",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "math.*",
            "file.*",
            "database.*",
            "http.*",
            "ai.*",
            "notify.*",
            "flow.*",
            # NO browser.* - data can't control browser
        ],
        can_receive_from=["*"],  # Can receive from anything
        description="Data modules transform data - cannot output to browser"
    ),

    "array": ConnectionRule(
        category="array",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "file.*",
            "database.*",
            "http.*",
            "ai.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="Array operations - cannot output to browser"
    ),

    "object": ConnectionRule(
        category="object",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "file.*",
            "database.*",
            "http.*",
            "ai.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="Object operations - cannot output to browser"
    ),

    "string": ConnectionRule(
        category="string",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "file.*",
            "database.*",
            "http.*",
            "ai.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="String operations - cannot output to browser"
    ),

    "math": ConnectionRule(
        category="math",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "math.*",
            "file.*",
            "database.*",
            "http.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="Math operations - cannot output to browser"
    ),

    "datetime": ConnectionRule(
        category="datetime",
        can_connect_to=[
            "data.*",
            "string.*",
            "file.*",
            "database.*",
            "http.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="DateTime operations - cannot output to browser"
    ),

    # =========================================================================
    # File System - Can receive from anywhere, limited output
    # =========================================================================
    "file": ConnectionRule(
        category="file",
        can_connect_to=[
            "file.*",       # Chain file operations
            "data.*",       # Process file content
            "pdf.*",        # PDF processing
            "image.*",      # Image processing
            "ai.*",         # AI analysis
            "notify.*",     # Send file
            "flow.*",       # Flow control
        ],
        can_receive_from=["*"],  # Any module can trigger file ops
        description="File operations - wide input, structured output"
    ),

    # =========================================================================
    # Image Processing - Similar to file
    # =========================================================================
    "image": ConnectionRule(
        category="image",
        can_connect_to=[
            "image.*",
            "file.*",
            "ai.*",
            "data.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=[
            "file.*",
            "image.*",
            "browser.*",  # Download image
            "http.*",
            "flow.*",
            "start",
        ],
        description="Image processing modules"
    ),

    # =========================================================================
    # Database - Structured data in/out
    # =========================================================================
    "database": ConnectionRule(
        category="database",
        can_connect_to=[
            "database.*",   # Chain queries
            "data.*",       # Transform results
            "array.*",      # Array operations
            "object.*",     # Object operations
            "file.*",       # Export to file
            "http.*",       # Send to API
            "notify.*",     # Notify
            "flow.*",       # Flow control
        ],
        can_receive_from=[
            "data.*",
            "array.*",
            "object.*",
            "file.*",
            "http.*",
            "flow.*",
            "start",
        ],
        description="Database modules - structured data only, no browser"
    ),

    # =========================================================================
    # API / HTTP - Data exchange, NO browser
    # =========================================================================
    "api": ConnectionRule(
        category="api",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "file.*",
            "database.*",
            "http.*",
            "ai.*",
            "notify.*",
            "flow.*",
            # NO browser.* - API responses don't control browser
        ],
        can_receive_from=["*"],
        description="API modules - data exchange, cannot output to browser"
    ),

    "http": ConnectionRule(
        category="http",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "file.*",
            "database.*",
            "http.*",
            "ai.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="HTTP modules - data exchange, cannot output to browser"
    ),

    # =========================================================================
    # AI / ML - STRICT: Cannot output to browser
    # =========================================================================
    "ai": ConnectionRule(
        category="ai",
        can_connect_to=[
            "data.*",
            "string.*",
            "array.*",
            "object.*",
            "file.*",
            "database.*",
            "http.*",
            "notify.*",
            "flow.*",
            "ai.*",
            # NO browser.* - AI cannot control browser
        ],
        can_receive_from=[
            "data.*",
            "string.*",
            "array.*",
            "object.*",
            "file.*",
            "image.*",
            "http.*",
            "database.*",
            "ai.*",
            "flow.*",
            "start",
            # Can receive browser data via element.get_* but not browser control
        ],
        description="AI modules process data - CANNOT output to browser"
    ),

    "llm": ConnectionRule(
        category="llm",
        can_connect_to=[
            "data.*",
            "string.*",
            "file.*",
            "http.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=["*"],
        description="LLM modules - text in, text out"
    ),

    # =========================================================================
    # Analysis - Can receive from browser, cannot output to browser
    # =========================================================================
    "analysis": ConnectionRule(
        category="analysis",
        can_connect_to=[
            "data.*",
            "array.*",
            "object.*",
            "string.*",
            "file.*",
            "database.*",
            "ai.*",
            "notify.*",
            "flow.*",
            # NO browser.* - analysis results don't control browser
        ],
        can_receive_from=[
            "browser.*",    # Analyze page
            "element.*",    # Analyze element
            "file.*",       # Analyze file
            "data.*",       # Analyze data
            "http.*",       # Analyze response
            "flow.*",
            "start",
        ],
        description="Analysis modules - can receive from browser, cannot output to browser"
    ),

    # =========================================================================
    # Document Processing - File-based
    # =========================================================================
    "document": ConnectionRule(
        category="document",
        can_connect_to=[
            "pdf.*",
            "data.*",
            "string.*",
            "file.*",
            "ai.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=[
            "file.*",
            "browser.*",    # Download from browser
            "http.*",       # Receive from API
            "pdf.*",
            "flow.*",
            "start",
        ],
        description="Document modules - file processing"
    ),

    # =========================================================================
    # Notify - Notification output endpoint
    # =========================================================================
    "notify": ConnectionRule(
        category="notify",
        can_connect_to=["notify.*", "data.*", "flow.*", "end"],
        can_receive_from=["*"],
        description="Notification modules - output endpoint"
    ),

    # =========================================================================
    # Notification / Output - End point, can receive from anywhere
    # =========================================================================
    "notification": ConnectionRule(
        category="notification",
        can_connect_to=[
            "notify.*",       # Chain notifications
            "data.*",          # Log response
            "flow.*",          # Continue after
            "end",             # End workflow
        ],
        can_receive_from=["*"],  # Any module can trigger notification
        description="Notification modules - output endpoint"
    ),

    "communication": ConnectionRule(
        category="communication",
        can_connect_to=[
            "notify.*",
            "email.*",
            "slack.*",
            "webhook.*",
            "data.*",
            "flow.*",
            "end",
        ],
        can_receive_from=["*"],
        description="Communication modules - output endpoint"
    ),

    # =========================================================================
    # Testing - Can connect to flow control
    # =========================================================================
    "test": ConnectionRule(
        category="test",
        can_connect_to=[
            "test.*",
            "flow.*",
            "notify.*",
            "data.*",
        ],
        can_receive_from=["*"],
        description="Test modules - assertion and validation"
    ),

    "testing": ConnectionRule(
        category="testing",
        can_connect_to=[
            "testing.*",
            "test.*",
            "flow.*",
            "notify.*",
            "data.*",
        ],
        can_receive_from=["*"],
        description="Testing modules - assertion and validation"
    ),

    # =========================================================================
    # Utility / Meta - Mostly universal but no browser output
    # =========================================================================
    "utility": ConnectionRule(
        category="utility",
        can_connect_to=[
            "data.*",
            "string.*",
            "file.*",
            "http.*",
            "notify.*",
            "flow.*",
            "utility.*",
        ],
        can_receive_from=["*"],
        description="Utility modules - helpers"
    ),

    "meta": ConnectionRule(
        category="meta",
        can_connect_to=["*"],
        can_receive_from=["*"],
        description="Meta modules are universal (logging, debug, etc.)"
    ),

    # =========================================================================
    # HuggingFace - AI models, no browser output
    # =========================================================================
    "huggingface": ConnectionRule(
        category="huggingface",
        can_connect_to=[
            "data.*",
            "string.*",
            "array.*",
            "object.*",
            "file.*",
            "ai.*",
            "huggingface.*",
            "notify.*",
            "flow.*",
        ],
        can_receive_from=[
            "data.*",
            "string.*",
            "file.*",
            "image.*",
            "http.*",
            "huggingface.*",
            "flow.*",
            "start",
        ],
        description="HuggingFace AI models - no browser interaction"
    ),

    # =========================================================================
    # Composite modules - Depend on their internal implementation
    # =========================================================================
    "composite": ConnectionRule(
        category="composite",
        can_connect_to=["*"],
        can_receive_from=["*"],
        description="Composite modules depend on their internal implementation"
    ),

    # =========================================================================
    # Shell/Process - System operations, restricted
    # =========================================================================
    "shell": ConnectionRule(
        category="shell",
        can_connect_to=[
            "data.*",
            "string.*",
            "file.*",
            "flow.*",
            "test.*",
        ],
        can_receive_from=[
            "data.*",
            "string.*",
            "file.*",
            "flow.*",
            "start",
        ],
        description="Shell operations - system commands"
    ),

    "process": ConnectionRule(
        category="process",
        can_connect_to=[
            "process.*",
            "port.*",
            "data.*",
            "flow.*",
            "test.*",
        ],
        can_receive_from=[
            "process.*",
            "flow.*",
            "start",
        ],
        description="Process management"
    ),

    "port": ConnectionRule(
        category="port",
        can_connect_to=[
            "browser.*",  # Port ready, launch browser
            "http.*",
            "flow.*",
            "test.*",
        ],
        can_receive_from=[
            "process.*",
            "flow.*",
            "start",
        ],
        description="Port operations - typically before browser/http"
    ),
}



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Google Calendar Create Event Module
Create a new event in Google Calendar using the Calendar API with OAuth2 and aiohttp.
"""

import logging
from typing import Any, Dict, List

from ....registry import register_module
from ....schema import compose
from ....schema.builders import field
from ....schema.constants import FieldGroup
from ....errors import ValidationError, ModuleError

logger = logging.getLogger(__name__)

CALENDAR_EVENTS_URL = 'https://www.googleapis.com/calendar/v3/calendars/primary/events'


@register_module(
    module_id='google.calendar.create_event',
    version='1.0.0',
    category='cloud',
    tags=['cloud', 'google', 'calendar', 'event', 'create', 'schedule'],
    label='Calendar Create Event',
    label_key='modules.google.calendar.create_event.label',
    description='Create a new event in Google Calendar',
    description_key='modules.google.calendar.create_event.description',
    icon='Calendar',
    color='#4285F4',
    input_types=['object', 'string'],
    output_types=['object'],
    can_receive_from=['*'],
    can_connect_to=['*'],
    retryable=True,
    max_retries=2,
    concurrent_safe=True,
    timeout_ms=30000,
    requires_credentials=True,
    handles_sensitive_data=False,
    required_permissions=['cloud.calendar'],
    params_schema=compose(
        field('access_token', type='string', label='Access Token', required=True,
              group=FieldGroup.CONNECTION,
              description='Google OAuth2 access token with Calendar write scope',
              format='password'),
        field('summary', type='string', label='Event Title', required=True,
              group=FieldGroup.BASIC,
              description='Title of the calendar event',
              placeholder='Team Meeting'),
        field('start_time', type='string', label='Start Time', required=True,
              group=FieldGroup.BASIC,
              description='Event start time in ISO 8601 format',
              placeholder='2026-03-01T10:00:00'),
        field('end_time', type='string', label='End Time', required=True,
              group=FieldGroup.BASIC,
              description='Event end time in ISO 8601 format',
              placeholder='2026-03-01T11:00:00'),
        field('description', type='string', label='Description',
              group=FieldGroup.OPTIONS,
              description='Detailed description of the event',
              placeholder='Discuss project milestones', format='multiline'),
        field('location', type='string', label='Location',
              group=FieldGroup.OPTIONS,
              description='Event location or meeting link',
              placeholder='Conference Room A'),
        field('attendees', type='string', label='Attendees',
              group=FieldGroup.OPTIONS,
              description='Comma-separated list of attendee email addresses',
              placeholder='alice@example.com, bob@example.com'),
        field('timezone', type='string', label='Timezone',
              group=FieldGroup.OPTIONS,
              description='Timezone for the event (IANA timezone)',
              default='UTC', placeholder='America/New_York'),
    ),
    output_schema={
        'event_id': {'type': 'string', 'description': 'Created event ID', 'description_key': 'modules.google.calendar.create_event.output.event_id.description'},
        'summary': {'type': 'string', 'description': 'Event title', 'description_key': 'modules.google.calendar.create_event.output.summary.description'},
        'start': {'type': 'string', 'description': 'Event start time', 'description_key': 'modules.google.calendar.create_event.output.start.description'},
        'end': {'type': 'string', 'description': 'Event end time', 'description_key': 'modules.google.calendar.create_event.output.end.description'},
        'html_link': {'type': 'string', 'description': 'Link to view the event in Google Calendar', 'description_key': 'modules.google.calendar.create_event.output.html_link.description'},
    },
    examples=[
        {
            'title': 'Create a meeting event',
            'params': {
                'access_token': '<oauth2-token>',
                'summary': 'Sprint Planning',
                'start_time': '2026-03-01T10:00:00',
                'end_time': '2026-03-01T11:00:00',
                'attendees': 'alice@example.com, bob@example.com',
                'timezone': 'America/New_York',
            },
        },
    ],
    author='Flyto Team',
    license='MIT',
)
async def google_calendar_create_event(context: Dict[str, Any]) -> Dict[str, Any]:
    """Create a Google Calendar event."""
    params = context.get('params', {})

    access_token = params.get('access_token')
    summary = params.get('summary')
    start_time = params.get('start_time')
    end_time = params.get('end_time')

    if not access_token:
        raise ValidationError('Access token is required', field='access_token')
    if not summary:
        raise ValidationError('Event title (summary) is required', field='summary')
    if not start_time:
        raise ValidationError('Start time is required', field='start_time')
    if not end_time:
        raise ValidationError('End time is required', field='end_time')

    description = params.get('description', '')
    location = params.get('location', '')
    attendees_str = params.get('attendees', '')
    timezone = params.get('timezone', 'UTC')

    # Build attendee list from comma-separated string
    attendee_list: List[Dict[str, str]] = []
    if attendees_str:
        for email in attendees_str.split(','):
            email = email.strip()
            if email:
                attendee_list.append({'email': email})

    # Build event payload
    event_body: Dict[str, Any] = {
        'summary': summary,
        'start': {
            'dateTime': start_time,
            'timeZone': timezone,
        },
        'end': {
            'dateTime': end_time,
            'timeZone': timezone,
        },
    }
    if description:
        event_body['description'] = description
    if location:
        event_body['location'] = location
    if attendee_list:
        event_body['attendees'] = attendee_list

    try:
        import aiohttp
    except ImportError:
        raise ModuleError(
            'aiohttp package is required. Install with: pip install aiohttp'
        )

    headers = {
        'Authorization': f'Bearer {access_token}',
        'Content-Type': 'application/json',
    }

    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                CALENDAR_EVENTS_URL,
                json=event_body,
                headers=headers,
                timeout=aiohttp.ClientTimeout(total=25),
            ) as resp:
                resp_data = await resp.json()
                if resp.status != 200:
                    error_msg = resp_data.get('error', {}).get('message', str(resp_data))
                    raise ModuleError(
                        f'Google Calendar API error (HTTP {resp.status}): {error_msg}'
                    )
    except aiohttp.ClientError as exc:
        raise ModuleError(f'Google Calendar API request failed: {exc}')

    # Extract start/end from response
    start_info = resp_data.get('start', {})
    end_info = resp_data.get('end', {})
    result_start = start_info.get('dateTime', start_info.get('date', ''))
    result_end = end_info.get('dateTime', end_info.get('date', ''))

    return {
        'ok': True,
        'data': {
            'event_id': resp_data.get('id', ''),
            'summary': resp_data.get('summary', ''),
            'start': result_start,
            'end': result_end,
            'html_link': resp_data.get('htmlLink', ''),
        },
    }



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Metering Module

Tracks usage and billing for module/plugin invocations.
"""

from .tracker import (
    MeteringTracker,
    MeteringRecord,
    MeteringConfig,
    CostClass,
    get_metering_tracker,
)

__all__ = [
    "MeteringTracker",
    "MeteringRecord",
    "MeteringConfig",
    "CostClass",
    "get_metering_tracker",
]



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Schema Auditor - Main audit logic for module schemas

Scans all registered modules and identifies quality issues in their
params_schema and output_schema definitions.
"""
from __future__ import annotations
from dataclasses import dataclass, field as dataclass_field
from typing import Any, Dict, List, Optional, Set
from collections import defaultdict

from .standards import (
    QualityLevel,
    REQUIRED_PARAM_FIELDS,
    RECOMMENDED_PARAM_FIELDS,
    STRING_PARAM_FIELDS,
    REQUIRED_OUTPUT_FIELDS,
    TYPES_REQUIRING_PLACEHOLDER,
    DEFAULT_PLACEHOLDERS,
    DEFAULT_DESCRIPTIONS,
    DEFAULT_DESCRIPTIONS_BY_KEY,
    DEFAULT_PLACEHOLDERS_BY_KEY,
    FLOW_CONTROL_MODULES,
)


@dataclass
class FieldIssue:
    """Represents a single schema field issue."""
    module_id: str
    field_name: str
    issue_type: str  # 'missing_description', 'missing_placeholder', 'missing_label', etc.
    level: QualityLevel
    message: str
    suggested_fix: Optional[str] = None
    current_value: Optional[Any] = None


@dataclass
class ModuleAuditResult:
    """Audit result for a single module."""
    module_id: str
    category: str
    issues: List[FieldIssue] = dataclass_field(default_factory=list)
    params_schema: Dict[str, Any] = dataclass_field(default_factory=dict)
    output_schema: Dict[str, Any] = dataclass_field(default_factory=dict)
    input_types: List[str] = dataclass_field(default_factory=list)
    output_types: List[str] = dataclass_field(default_factory=list)
    uses_compose: bool = False

    @property
    def critical_count(self) -> int:
        return sum(1 for i in self.issues if i.level == QualityLevel.CRITICAL)

    @property
    def warning_count(self) -> int:
        return sum(1 for i in self.issues if i.level == QualityLevel.WARNING)

    @property
    def info_count(self) -> int:
        return sum(1 for i in self.issues if i.level == QualityLevel.INFO)

    @property
    def has_issues(self) -> bool:
        return len(self.issues) > 0


@dataclass
class AuditResult:
    """Complete audit result across all modules."""
    modules: List[ModuleAuditResult] = dataclass_field(default_factory=list)
    total_modules: int = 0
    modules_with_issues: int = 0
    total_issues: int = 0
    critical_issues: int = 0
    warning_issues: int = 0
    info_issues: int = 0
    by_category: Dict[str, List[ModuleAuditResult]] = dataclass_field(default_factory=dict)
    by_issue_type: Dict[str, List[FieldIssue]] = dataclass_field(default_factory=dict)

    def add_module_result(self, result: ModuleAuditResult) -> None:
        """Add a module audit result."""
        self.modules.append(result)
        self.total_modules += 1

        if result.has_issues:
            self.modules_with_issues += 1
            self.total_issues += len(result.issues)
            self.critical_issues += result.critical_count
            self.warning_issues += result.warning_count
            self.info_issues += result.info_count

        # Index by category
        if result.category not in self.by_category:
            self.by_category[result.category] = []
        self.by_category[result.category].append(result)

        # Index by issue type
        for issue in result.issues:
            if issue.issue_type not in self.by_issue_type:
                self.by_issue_type[issue.issue_type] = []
            self.by_issue_type[issue.issue_type].append(issue)


class SchemaAuditor:
    """
    Audits module schemas for quality issues.

    Usage:
        auditor = SchemaAuditor()
        result = auditor.audit_all()

        # Or audit a single module
        module_result = auditor.audit_module(module_id, metadata)
    """

    def __init__(self, strict: bool = False):
        """
        Initialize the auditor.

        Args:
            strict: If True, treat warnings as critical issues
        """
        self.strict = strict

    def audit_all(self) -> AuditResult:
        """
        Audit all registered modules.

        Returns:
            AuditResult with all findings
        """
        from ..registry.core import ModuleRegistry

        result = AuditResult()

        # Get all module metadata
        all_metadata = ModuleRegistry.get_all_metadata(filter_by_stability=False)

        for module_id, metadata in all_metadata.items():
            module_result = self.audit_module(module_id, metadata)
            result.add_module_result(module_result)

        return result

    def audit_module(self, module_id: str, metadata: Dict[str, Any]) -> ModuleAuditResult:
        """
        Audit a single module's schema.

        Args:
            module_id: Module identifier
            metadata: Module metadata from registry

        Returns:
            ModuleAuditResult with findings
        """
        result = ModuleAuditResult(
            module_id=module_id,
            category=metadata.get('category', module_id.split('.')[0]),
            params_schema=metadata.get('params_schema', {}),
            output_schema=metadata.get('output_schema', {}),
            input_types=metadata.get('input_types', []),
            output_types=metadata.get('output_types', []),
        )

        # Audit params_schema
        self._audit_params_schema(result)

        # Audit output_schema
        self._audit_output_schema(result)

        # Audit input/output types
        self._audit_types(result)

        return result

    def _audit_params_schema(self, result: ModuleAuditResult) -> None:
        """Audit params_schema for quality issues."""
        params_schema = result.params_schema

        for field_name, field_def in params_schema.items():
            # Skip special fields
            if field_name.startswith('__'):
                continue

            if not isinstance(field_def, dict):
                result.issues.append(FieldIssue(
                    module_id=result.module_id,
                    field_name=field_name,
                    issue_type='invalid_field_def',
                    level=QualityLevel.CRITICAL,
                    message=f"Field definition is not a dict: {type(field_def)}",
                ))
                continue

            field_type = field_def.get('type', 'string')
            # Handle case where type might be a list
            if isinstance(field_type, list):
                field_type = field_type[0] if field_type else 'string'

            # Check for missing description
            if 'description' not in field_def:
                suggested = self._suggest_description(field_name, field_type)
                result.issues.append(FieldIssue(
                    module_id=result.module_id,
                    field_name=field_name,
                    issue_type='missing_description',
                    level=QualityLevel.WARNING,
                    message=f"Missing description for param '{field_name}'",
                    suggested_fix=suggested,
                ))

            # Check for missing label
            if 'label' not in field_def:
                suggested = self._suggest_label(field_name)
                result.issues.append(FieldIssue(
                    module_id=result.module_id,
                    field_name=field_name,
                    issue_type='missing_label',
                    level=QualityLevel.WARNING,
                    message=f"Missing label for param '{field_name}'",
                    suggested_fix=suggested,
                ))

            # Check for missing placeholder (string types)
            if field_type in TYPES_REQUIRING_PLACEHOLDER or field_type == 'string':
                if 'placeholder' not in field_def:
                    suggested = self._suggest_placeholder(field_name, field_type)
                    result.issues.append(FieldIssue(
                        module_id=result.module_id,
                        field_name=field_name,
                        issue_type='missing_placeholder',
                        level=QualityLevel.WARNING,
                        message=f"Missing placeholder for string param '{field_name}'",
                        suggested_fix=suggested,
                    ))

    def _audit_output_schema(self, result: ModuleAuditResult) -> None:
        """Audit output_schema for quality issues."""
        output_schema = result.output_schema

        for field_name, field_def in output_schema.items():
            # Skip special fields
            if field_name.startswith('__'):
                continue

            if not isinstance(field_def, dict):
                continue

            # Check for missing description in output
            if 'description' not in field_def:
                result.issues.append(FieldIssue(
                    module_id=result.module_id,
                    field_name=f"output.{field_name}",
                    issue_type='output_missing_description',
                    level=QualityLevel.INFO,
                    message=f"Missing description for output field '{field_name}'",
                    suggested_fix=self._suggest_description(field_name, field_def.get('type', 'any')),
                ))

    def _audit_types(self, result: ModuleAuditResult) -> None:
        """Audit input/output types."""
        # Check for missing input_types
        if not result.input_types:
            # Flow control modules are expected to have special handling
            if result.module_id in FLOW_CONTROL_MODULES:
                level = QualityLevel.WARNING
            else:
                level = QualityLevel.INFO

            result.issues.append(FieldIssue(
                module_id=result.module_id,
                field_name='input_types',
                issue_type='missing_input_types',
                level=level,
                message=f"Missing input_types for module",
                suggested_fix=self._suggest_input_types(result.module_id, result.category),
            ))

        # Check for missing output_types
        if not result.output_types:
            if result.module_id in FLOW_CONTROL_MODULES:
                level = QualityLevel.WARNING
            else:
                level = QualityLevel.INFO

            result.issues.append(FieldIssue(
                module_id=result.module_id,
                field_name='output_types',
                issue_type='missing_output_types',
                level=level,
                message=f"Missing output_types for module",
                suggested_fix=self._suggest_output_types(result.module_id, result.category),
            ))

    def _suggest_description(self, field_name: str, field_type: str) -> str:
        """Suggest a description for a field."""
        # First check by field name
        if field_name in DEFAULT_DESCRIPTIONS_BY_KEY:
            return DEFAULT_DESCRIPTIONS_BY_KEY[field_name]

        # Then check by type
        if field_type in DEFAULT_DESCRIPTIONS:
            return DEFAULT_DESCRIPTIONS[field_type]

        # Generate from field name
        return f"{field_name.replace('_', ' ').title()} value"

    def _suggest_label(self, field_name: str) -> str:
        """Suggest a label for a field."""
        return field_name.replace('_', ' ').title()

    def _suggest_placeholder(self, field_name: str, field_type: str) -> str:
        """Suggest a placeholder for a field."""
        # First check by field name
        if field_name in DEFAULT_PLACEHOLDERS_BY_KEY:
            return DEFAULT_PLACEHOLDERS_BY_KEY[field_name]

        # Then check by type
        if field_type in DEFAULT_PLACEHOLDERS:
            return DEFAULT_PLACEHOLDERS[field_type]

        return 'Enter value...'

    def _suggest_input_types(self, module_id: str, category: str) -> str:
        """Suggest input types based on module/category."""
        suggestions = {
            'string': "['string']",
            'array': "['array']",
            'object': "['object']",
            'math': "['number']",
            'browser': "['browser']",
            'file': "['string']",  # file path
            'http': "['string']",  # url
            'data': "['any']",
            'flow': "['control']",
            'error': "['any']",
        }
        return suggestions.get(category, "['any']")

    def _suggest_output_types(self, module_id: str, category: str) -> str:
        """Suggest output types based on module/category."""
        suggestions = {
            'string': "['string']",
            'array': "['array']",
            'object': "['object']",
            'math': "['number']",
            'browser': "['browser']",
            'file': "['string']",
            'http': "['object']",
            'data': "['any']",
            'flow': "['control']",
            'error': "['any']",
        }
        return suggestions.get(category, "['any']")


def run_audit(strict: bool = False) -> AuditResult:
    """
    Convenience function to run a full audit.

    Args:
        strict: If True, treat warnings as critical

    Returns:
        AuditResult with all findings
    """
    auditor = SchemaAuditor(strict=strict)
    return auditor.audit_all()



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Test Report Generator Module

Generate test execution reports.
"""

import logging
from typing import Any, Dict

from ...registry import register_module

logger = logging.getLogger(__name__)


@register_module(
    module_id='testing.report.generate',
    version='1.0.0',
    category='atomic',
    subcategory='testing',
    tags=['testing', 'report', 'documentation', 'atomic', 'path_restricted'],
    label='Generate Report',
    label_key='modules.testing.report.generate.label',
    description='Generate test execution report',
    description_key='modules.testing.report.generate.description',
    icon='FileText',
    color='#6366F1',

    input_types=['object', 'array'],
    output_types=['object', 'string'],
    can_receive_from=['testing.*', 'data.*', 'flow.*'],
    can_connect_to=['notify.*', 'file.*', 'data.*', 'flow.*', 'end'],

    timeout_ms=60000,
    retryable=False,

    params_schema={
        'results': {
            'type': 'object',
            'label': 'Test Results',
            'required': True,
            'description': 'Results data',
        },
        'format': {
            'type': 'string',
            'label': 'Report Format',
            'default': 'json',
            'options': ['json', 'html', 'markdown', 'junit'],
            'placeholder': 'format-string',
            'description': 'Format',
        },
        'title': {
            'type': 'string',
            'label': 'Report Title',
            'default': 'Test Report',
            'placeholder': 'Title',
            'description': 'Title',
        }
    },
    output_schema={
        'ok': {'type': 'boolean', 'description': 'Whether the operation succeeded',
                'description_key': 'modules.testing.report.generate.output.ok.description'},
        'report': {'type': 'string', 'description': 'The report',
                'description_key': 'modules.testing.report.generate.output.report.description'},
        'format': {'type': 'string', 'description': 'The format',
                'description_key': 'modules.testing.report.generate.output.format.description'},
        'summary': {'type': 'object', 'description': 'The summary',
                'description_key': 'modules.testing.report.generate.output.summary.description'}
    }
)
async def testing_report_generate(context: Dict[str, Any]) -> Dict[str, Any]:
    """Generate test report"""
    import json

    params = context['params']
    results = params.get('results', {})
    format_type = params.get('format', 'json')
    title = params.get('title', 'Test Report')

    summary = {
        'title': title,
        'total': results.get('total', 0),
        'passed': results.get('passed', 0),
        'failed': results.get('failed', 0)
    }

    if format_type == 'json':
        report = json.dumps({'title': title, 'results': results}, indent=2)
    elif format_type == 'markdown':
        report = f"# {title}\n\n"
        report += f"- Total: {summary['total']}\n"
        report += f"- Passed: {summary['passed']}\n"
        report += f"- Failed: {summary['failed']}\n"
    else:
        report = json.dumps(results)

    return {
        'ok': True,
        'report': report,
        'format': format_type,
        'summary': summary
    }



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Notification Composite Modules

High-level notification workflows combining multiple channels.
"""
from .multi_channel_alert import MultiChannelAlert
from .scheduled_report import ScheduledReport

__all__ = [
    'MultiChannelAlert',
    'ScheduledReport',
]



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Jira Search Issues Module

Search issues using JQL query.
"""

import os
from typing import Any, Dict

from ....base import BaseModule
from ....registry import register_module
from ..integration import JiraIntegration


@register_module(
    module_id="integration.jira.search_issues",
    can_connect_to=['*'],
    can_receive_from=['*'],
    version="1.0.0",
    category="integration",
    tags=["integration", "jira", "search", "jql", "ssrf_protected"],
    label="Search Jira Issues",
    label_key="modules.integration.jira.search_issues.label",
    description="Search issues using JQL query",
    description_key="modules.integration.jira.search_issues.description",
    icon="Search",
    color="#0052CC",
    input_types=["any"],
    output_types=["any"],
    timeout_ms=60000,
    retryable=True,
    concurrent_safe=True,
    requires_credentials=True,
    credential_keys=['JIRA_TOKEN', 'JIRA_EMAIL'],
    params_schema={
        "domain": {
            "type": "string",
            "label": "Jira Domain",
            "description": "Your Jira Cloud domain (e.g., your-domain.atlassian.net)",
            "description_key": "modules.integration.jira.search_issues.params.domain.description",
            "placeholder": "${env.JIRA_DOMAIN}",
            "required": True,
        },
        "jql": {
            "type": "string",
            "label": "JQL Query",
            "description": "JQL search query",
                "description_key": "modules.integration.jira.search_issues.params.jql.description",
            "placeholder": "project = PROJ AND status = Open",
            "required": True,
        },
        "max_results": {
            "type": "number",
            "label": "Max Results",
            "default": 50,
            "min": 1,
            "max": 1000,
            "required": False,
        },
        "email": {
            "type": "string",
            "placeholder": "${env.JIRA_EMAIL}",
            "required": False,
        },
        "api_token": {
            "type": "string",
            "placeholder": "${env.JIRA_API_TOKEN}",
            "required": False,
            "sensitive": True,
        },
    },
    output_schema={
        "ok": {"type": "boolean", "description": "Whether the operation was successful"},
        "issues": {"type": "array", "description": "List of matching issues"},
        "total": {"type": "number", "description": "Total number of matching issues"},
    },
    author="Flyto Team",
    license="MIT",
)
class JiraSearchIssuesModule(BaseModule):
    """Search Jira issues module."""

    module_name = "Search Jira Issues"
    module_description = "Search issues using JQL"

    def validate_params(self) -> None:
        if not self.params.get("domain"):
            raise ValueError("Jira domain required")
        if not self.params.get("jql"):
            raise ValueError("JQL query required")

        self.domain = self.params["domain"]
        self.jql = self.params["jql"]
        self.max_results = self.params.get("max_results", 50)
        self.email = self.params.get("email") or os.getenv("JIRA_EMAIL")
        self.api_token = self.params.get("api_token") or os.getenv("JIRA_API_TOKEN")

    async def execute(self) -> Dict[str, Any]:
        async with JiraIntegration(
            domain=self.domain,
            email=self.email,
            api_token=self.api_token,
        ) as jira:
            response = await jira.search_issues(
                jql=self.jql,
                max_results=self.max_results,
                fields=["summary", "status", "priority", "assignee", "created", "updated"],
            )

            if response.ok:
                data = response.data
                issues = [
                    {
                        "key": issue.get("key"),
                        "summary": issue.get("fields", {}).get("summary"),
                        "status": issue.get("fields", {}).get("status", {}).get("name"),
                        "priority": issue.get("fields", {}).get("priority", {}).get("name"),
                        "assignee": issue.get("fields", {}).get("assignee", {}).get("displayName"),
                        "url": f"https://{self.domain}/browse/{issue.get('key')}",
                    }
                    for issue in data.get("issues", [])
                ]
                return {
                    "ok": True,
                    "issues": issues,
                    "total": data.get("total", 0),
                }
            else:
                return {
                    "ok": False,
                    "error": response.error,
                }



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Connection Rules Models

Data classes and enums for connection rules.
"""

from dataclasses import dataclass
from enum import Enum
from typing import List


class ConnectionCategory(str, Enum):
    """Categories for connection rule grouping"""
    BROWSER = "browser"
    FLOW = "flow"
    DATA = "data"
    FILE = "file"
    DATABASE = "database"
    API = "api"
    AI = "ai"
    DOCUMENT = "document"
    NOTIFICATION = "notification"
    ANALYSIS = "analysis"
    UTILITY = "utility"
    ANY = "*"


@dataclass
class ConnectionRule:
    """
    Connection rule for a category of modules.

    Attributes:
        category: Module category (e.g., "browser")
        can_connect_to: List of patterns for allowed targets
        can_receive_from: List of patterns for allowed sources
        description: Human-readable rule description
    """
    category: str
    can_connect_to: List[str]
    can_receive_from: List[str]
    description: str = ""



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Edge-Based Loop Execution

Handles loop execution using output port routing (Workflow Spec v1.2).
Returns events ('iterate'/'done') for workflow engine to handle routing.

Supports two modes:
- Repeat mode: Execute N times (times parameter only)
- ForEach mode: Iterate over items array, sets loop.item and loop.index scope
"""
from typing import Any, Dict, List, Optional

ITERATION_PREFIX = '__loop_iteration_'


async def execute_edge_mode(
    target: str,
    times: int,
    context: Dict[str, Any],
    items: Optional[List[Any]] = None,
    item_var: str = 'item',
    index_var: str = 'index',
) -> Dict[str, Any]:
    """
    Edge-based loop: return event for workflow engine routing.

    Two modes:
    1. Repeat mode (items=None): Just repeat N times
    2. ForEach mode (items=[...]): Iterate over items, set loop scope

    Args:
        target: Target step ID to jump back to (deprecated in v2.0)
        times: Number of times to execute (used if items is None)
        context: Execution context for tracking iteration state
        items: Optional array to iterate over (ForEach mode)
        item_var: Variable name for current item (default: 'item')
        index_var: Variable name for current index (default: 'index')

    Returns:
        Dict with __event__ (iterate/done) and __set_context for scope:
        - iterate: Continue looping, sets loop.item and loop.index
        - done: Loop complete, clears loop scope
    """
    # Determine iteration count
    if items is not None:
        total_iterations = len(items)
    else:
        total_iterations = times

    # Use step_id or target as key to track iterations
    step_id = context.get('__current_step_id', target or 'loop')
    iteration_key = f"{ITERATION_PREFIX}{step_id}"
    current_index = context.get(iteration_key, 0)

    # Check if we've completed all iterations
    if current_index >= total_iterations:
        # Loop complete - emit 'done' event
        return {
            '__event__': 'done',
            'outputs': {
                'done': {
                    'iterations': current_index,
                    'status': 'completed'
                }
            },
            'iteration': current_index,
            'status': 'completed',
            'message': f"Loop completed after {current_index} iterations",
            '__set_context': {
                iteration_key: 0,  # Reset for next execution
                # Clear loop scope
                f'loop.{item_var}': None,
                f'loop.{index_var}': None,
                'loop.item': None,
                'loop.index': None,
            }
        }

    # Continue iterating - emit 'iterate' event
    set_context = {
        iteration_key: current_index + 1,  # Increment for next iteration
        f'loop.{index_var}': current_index,
        'loop.index': current_index,
    }

    # ForEach mode: set current item in scope
    if items is not None:
        current_item = items[current_index]
        set_context[f'loop.{item_var}'] = current_item
        set_context['loop.item'] = current_item

    response = {
        '__event__': 'iterate',
        'outputs': {
            'iterate': {
                'index': current_index,
                'remaining': total_iterations - current_index - 1,
                'total': total_iterations,
            }
        },
        'iteration': current_index,
        '__set_context': set_context,
    }

    # Add item to outputs if in ForEach mode
    if items is not None:
        response['outputs']['iterate']['item'] = items[current_index]

    # Legacy support: include next_step if target provided (deprecated)
    if target:
        response['next_step'] = target

    return response



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Data Processing Modules

Advanced data transformation and processing modules including:
- pipeline: Chain multiple transformations in a single step
"""

from .pipeline import DataPipelineModule
from . import json_parse
from . import json_stringify
from . import text_template
from . import xml_parse
from . import xml_generate
from . import yaml_parse
from . import yaml_generate

__all__ = [
    'DataPipelineModule',
]



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Slack Send Module
Send messages to Slack channels via webhook
"""
import logging
import os
from typing import Any, Dict, List, Optional

from ...registry import register_module
from ...schema import compose, presets


logger = logging.getLogger(__name__)


@register_module(
    module_id='slack.send',
    stability="beta",
    version='1.0.0',
    category='communication',
    subcategory='slack',
    tags=['slack', 'message', 'send', 'notification', 'webhook', 'ssrf_protected'],
    label='Send Slack Message',
    label_key='modules.slack.send.label',
    description='Send messages to Slack channels via incoming webhook',
    description_key='modules.slack.send.description',
    icon='MessageSquare',
    color='#4A154B',

    input_types=['text', 'object'],
    output_types=['object'],
    can_connect_to=['notify.*', 'flow.*', 'data.*', 'string.*', 'object.*'],
    can_receive_from=['*'],

    timeout_ms=30000,
    retryable=True,
    max_retries=3,
    concurrent_safe=True,

    requires_credentials=True,
    credential_keys=['API_KEY'],
    handles_sensitive_data=False,
    required_permissions=[],

    params_schema=compose(
        presets.SLACK_MESSAGE(),
        presets.SLACK_WEBHOOK_URL(),
        presets.SLACK_CHANNEL(),
        presets.SLACK_USERNAME(),
        presets.SLACK_ICON_EMOJI(),
        presets.SLACK_BLOCKS(),
        presets.SLACK_ATTACHMENTS(),
    ),
    output_schema={
        'sent': {
            'type': 'boolean',
            'description': 'Whether message was sent successfully'
        ,
                'description_key': 'modules.slack.send.output.sent.description'}
    },
    examples=[
        {
            'title': 'Send simple message',
            'title_key': 'modules.slack.send.examples.simple.title',
            'params': {
                'message': 'Hello from Flyto!'
            }
        },
        {
            'title': 'Send with formatting',
            'title_key': 'modules.slack.send.examples.formatted.title',
            'params': {
                'message': 'Task completed successfully',
                'username': 'Flyto Bot',
                'icon_emoji': ':white_check_mark:'
            }
        }
    ],
    author='Flyto Team',
    license='MIT'
)
async def slack_send(context: Dict[str, Any]) -> Dict[str, Any]:
    """Send message to Slack via webhook"""
    try:
        import aiohttp
    except ImportError:
        raise ImportError("aiohttp is required for slack.send. Install with: pip install aiohttp")

    params = context['params']
    message = params['message']
    webhook_url = params.get('webhook_url') or os.getenv('SLACK_WEBHOOK_URL')
    channel = params.get('channel')
    username = params.get('username')
    icon_emoji = params.get('icon_emoji')
    blocks = params.get('blocks')
    attachments = params.get('attachments')

    if not webhook_url:
        raise ValueError("Slack webhook URL not configured. Set SLACK_WEBHOOK_URL env or provide webhook_url param")

    payload = {'text': message}

    if channel:
        payload['channel'] = channel
    if username:
        payload['username'] = username
    if icon_emoji:
        payload['icon_emoji'] = icon_emoji
    if blocks:
        payload['blocks'] = blocks
    if attachments:
        payload['attachments'] = attachments

    async with aiohttp.ClientSession() as session:
        async with session.post(webhook_url, json=payload) as response:
            if response.status == 200:
                logger.info("Slack message sent successfully")
                return {
                    'ok': True,
                    'sent': True
                }
            else:
                text = await response.text()
                logger.error(f"Failed to send Slack message: {response.status} - {text}")
                raise RuntimeError(f"Slack API error: {response.status} - {text}")



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Replay Manager

Manages workflow replay operations.
"""

import copy
import json
import logging
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from .models import ReplayConfig, ReplayResult

logger = logging.getLogger(__name__)


class ReplayManager:
    """
    Manages workflow replay operations.

    Loads evidence from previous executions and enables re-execution
    from any point with optional context modifications.

    Usage:
        manager = ReplayManager(evidence_store)

        # Replay from specific step
        result = await manager.replay_from_step(
            execution_id="exec_123",
            step_id="step_5",
            modified_context={"user_id": "new_user"},
            workflow_executor=executor_func,
        )

        # Validate replay without executing
        validation = await manager.validate_replay(
            execution_id="exec_123",
            step_id="step_5",
        )
    """

    def __init__(
        self,
        evidence_path: Path,
        max_replay_history: int = 100,
    ):
        """
        Initialize replay manager.

        Args:
            evidence_path: Path to evidence storage
            max_replay_history: Maximum replay history to maintain
        """
        self.evidence_path = Path(evidence_path)
        self.max_replay_history = max_replay_history
        self._replay_history: List[ReplayResult] = []

    async def load_execution_state(
        self,
        execution_id: str,
        step_id: str,
    ) -> Optional[Dict[str, Any]]:
        """
        Load execution state at a specific step.

        Args:
            execution_id: Original execution ID
            step_id: Step to load state at

        Returns:
            Context state at that step, or None if not found
        """
        jsonl_path = self.evidence_path / execution_id / "evidence.jsonl"

        if not jsonl_path.exists():
            logger.warning(f"Evidence not found for {execution_id}")
            return None

        try:
            with open(jsonl_path, 'r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue

                    data = json.loads(line)
                    if data.get('step_id') == step_id:
                        return {
                            "context_before": data.get('context_before', {}),
                            "context_after": data.get('context_after', {}),
                            "step_index": data.get('step_index'),
                            "module_id": data.get('module_id'),
                            "status": data.get('status'),
                            "output": data.get('output', {}),
                        }

        except Exception as e:
            logger.error(f"Failed to load state for {execution_id}/{step_id}: {e}")

        return None

    async def load_execution_steps(
        self,
        execution_id: str,
    ) -> List[Dict[str, Any]]:
        """
        Load all steps from an execution.

        Args:
            execution_id: Execution ID

        Returns:
            List of step data in execution order
        """
        jsonl_path = self.evidence_path / execution_id / "evidence.jsonl"

        if not jsonl_path.exists():
            return []

        steps = []
        try:
            with open(jsonl_path, 'r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if line:
                        steps.append(json.loads(line))
        except Exception as e:
            logger.error(f"Failed to load steps for {execution_id}: {e}")

        return steps

    async def validate_replay(
        self,
        execution_id: str,
        step_id: str,
        config: Optional[ReplayConfig] = None,
    ) -> Dict[str, Any]:
        """
        Validate that a replay is possible.

        Args:
            execution_id: Execution to replay from
            step_id: Step to start from
            config: Optional replay configuration

        Returns:
            Validation result with issues if any
        """
        config = config or ReplayConfig(start_step_id=step_id)
        issues = []
        warnings = []

        # Check evidence exists
        exec_dir = self.evidence_path / execution_id
        if not exec_dir.exists():
            issues.append(f"Execution {execution_id} not found")
            return {
                "valid": False,
                "issues": issues,
                "warnings": warnings,
            }

        # Load steps
        steps = await self.load_execution_steps(execution_id)
        if not steps:
            issues.append("No step evidence found")
            return {
                "valid": False,
                "issues": issues,
                "warnings": warnings,
            }

        # Check start step exists
        step_ids = [s.get('step_id') for s in steps]
        if step_id not in step_ids:
            issues.append(f"Step {step_id} not found in execution")
            return {
                "valid": False,
                "issues": issues,
                "warnings": warnings,
            }

        # Check end step if specified
        if config.end_step_id and config.end_step_id not in step_ids:
            issues.append(f"End step {config.end_step_id} not found")

        # Check skip steps exist
        for skip_id in config.skip_steps:
            if skip_id not in step_ids:
                warnings.append(f"Skip step {skip_id} not found")

        # Load state at start step
        state = await self.load_execution_state(execution_id, step_id)
        if not state:
            issues.append(f"Could not load state at step {step_id}")

        # Validate modified context keys
        if state and config.modified_context:
            original_keys = set(state.get('context_before', {}).keys())
            for key in config.modified_context:
                if key not in original_keys:
                    warnings.append(f"Modified key '{key}' not in original context")

        return {
            "valid": len(issues) == 0,
            "issues": issues,
            "warnings": warnings,
            "step_count": len(steps),
            "start_step_index": step_ids.index(step_id) if step_id in step_ids else None,
            "available_steps": step_ids,
        }

    async def prepare_replay_context(
        self,
        execution_id: str,
        step_id: str,
        modified_context: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Prepare context for replay execution.

        Loads the context_before from the target step and applies
        any modifications.

        Args:
            execution_id: Original execution ID
            step_id: Step to start from
            modified_context: Context modifications to apply

        Returns:
            Prepared context for replay
        """
        state = await self.load_execution_state(execution_id, step_id)

        if not state:
            raise ValueError(f"Cannot load state for {execution_id}/{step_id}")

        # Start with context before the step
        context = copy.deepcopy(state.get('context_before', {}))

        # Apply modifications
        if modified_context:
            context.update(modified_context)

        return context

    async def replay_from_step(
        self,
        execution_id: str,
        step_id: str,
        workflow_executor: Callable,
        config: Optional[ReplayConfig] = None,
    ) -> ReplayResult:
        """
        Execute replay from a specific step.

        Args:
            execution_id: Original execution ID
            step_id: Step to start from
            workflow_executor: Async function to execute workflow
                Signature: async def executor(workflow, context, start_step, end_step) -> result
            config: Replay configuration

        Returns:
            ReplayResult with execution outcome
        """
        config = config or ReplayConfig(start_step_id=step_id)
        start_time = datetime.now()
        new_execution_id = f"replay_{uuid.uuid4().hex[:12]}"

        # Validate first
        validation = await self.validate_replay(execution_id, step_id, config)
        if not validation.get('valid'):
            return ReplayResult(
                ok=False,
                execution_id=new_execution_id,
                original_execution_id=execution_id,
                start_step=step_id,
                error=f"Validation failed: {validation.get('issues')}",
            )

        if config.dry_run:
            return ReplayResult(
                ok=True,
                execution_id=new_execution_id,
                original_execution_id=execution_id,
                start_step=step_id,
                context={"dry_run": True, "validation": validation},
            )

        try:
            # Prepare context
            context = await self.prepare_replay_context(
                execution_id,
                step_id,
                config.modified_context,
            )

            # Get workflow definition (stored in evidence metadata)
            workflow = await self._load_workflow_definition(execution_id)
            if not workflow:
                return ReplayResult(
                    ok=False,
                    execution_id=new_execution_id,
                    original_execution_id=execution_id,
                    start_step=step_id,
                    error="Workflow definition not found in evidence",
                )

            # Execute workflow from step
            result = await workflow_executor(
                workflow=workflow,
                context=context,
                start_step=step_id,
                end_step=config.end_step_id,
                skip_steps=config.skip_steps,
                breakpoints=config.breakpoints,
            )

            duration_ms = int((datetime.now() - start_time).total_seconds() * 1000)

            replay_result = ReplayResult(
                ok=result.get('ok', False),
                execution_id=new_execution_id,
                original_execution_id=execution_id,
                start_step=step_id,
                end_step=config.end_step_id,
                steps_executed=result.get('steps_executed', 0),
                steps_skipped=len(config.skip_steps),
                context=result.get('context', {}),
                error=result.get('error'),
                duration_ms=duration_ms,
            )

            # Record in history
            self._add_to_history(replay_result)

            return replay_result

        except Exception as e:
            logger.error(f"Replay failed: {e}")
            duration_ms = int((datetime.now() - start_time).total_seconds() * 1000)

            return ReplayResult(
                ok=False,
                execution_id=new_execution_id,
                original_execution_id=execution_id,
                start_step=step_id,
                error=str(e),
                duration_ms=duration_ms,
            )

    async def replay_single_step(
        self,
        execution_id: str,
        step_id: str,
        step_executor: Callable,
        modified_params: Optional[Dict[str, Any]] = None,
        modified_context: Optional[Dict[str, Any]] = None,
    ) -> ReplayResult:
        """
        Re-execute a single step.

        Useful for testing individual step behavior with different inputs.

        Args:
            execution_id: Original execution ID
            step_id: Step to re-execute
            step_executor: Async function to execute single step
                Signature: async def executor(module_id, params, context) -> result
            modified_params: Parameter modifications
            modified_context: Context modifications

        Returns:
            ReplayResult with step outcome
        """
        start_time = datetime.now()
        new_execution_id = f"replay_step_{uuid.uuid4().hex[:12]}"

        try:
            # Load step state
            state = await self.load_execution_state(execution_id, step_id)
            if not state:
                return ReplayResult(
                    ok=False,
                    execution_id=new_execution_id,
                    original_execution_id=execution_id,
                    start_step=step_id,
                    error=f"Step {step_id} not found",
                )

            # Prepare context
            context = copy.deepcopy(state.get('context_before', {}))
            if modified_context:
                context.update(modified_context)

            # Get step configuration from workflow
            workflow = await self._load_workflow_definition(execution_id)
            step_config = self._find_step_config(workflow, step_id)

            if not step_config:
                return ReplayResult(
                    ok=False,
                    execution_id=new_execution_id,
                    original_execution_id=execution_id,
                    start_step=step_id,
                    error=f"Step configuration not found for {step_id}",
                )

            # Merge parameters
            params = copy.deepcopy(step_config.get('params', {}))
            if modified_params:
                params.update(modified_params)

            # Execute step
            result = await step_executor(
                module_id=step_config.get('module'),
                params=params,
                context=context,
            )

            duration_ms = int((datetime.now() - start_time).total_seconds() * 1000)

            return ReplayResult(
                ok=result.get('ok', True),
                execution_id=new_execution_id,
                original_execution_id=execution_id,
                start_step=step_id,
                end_step=step_id,
                steps_executed=1,
                context=result.get('context', context),
                error=result.get('error'),
                duration_ms=duration_ms,
            )

        except Exception as e:
            logger.error(f"Single step replay failed: {e}")
            return ReplayResult(
                ok=False,
                execution_id=new_execution_id,
                original_execution_id=execution_id,
                start_step=step_id,
                error=str(e),
            )

    async def compare_replay(
        self,
        original_execution_id: str,
        replay_execution_id: str,
    ) -> Dict[str, Any]:
        """
        Compare original and replay execution results.

        Args:
            original_execution_id: Original execution ID
            replay_execution_id: Replay execution ID

        Returns:
            Comparison results
        """
        original_steps = await self.load_execution_steps(original_execution_id)
        replay_steps = await self.load_execution_steps(replay_execution_id)

        original_by_id = {s['step_id']: s for s in original_steps}
        replay_by_id = {s['step_id']: s for s in replay_steps}

        differences = []
        for step_id in set(original_by_id.keys()) | set(replay_by_id.keys()):
            orig = original_by_id.get(step_id)
            repl = replay_by_id.get(step_id)

            if not orig:
                differences.append({
                    "step_id": step_id,
                    "type": "added_in_replay",
                })
            elif not repl:
                differences.append({
                    "step_id": step_id,
                    "type": "not_in_replay",
                })
            else:
                # Compare outcomes
                if orig.get('status') != repl.get('status'):
                    differences.append({
                        "step_id": step_id,
                        "type": "status_changed",
                        "original": orig.get('status'),
                        "replay": repl.get('status'),
                    })

                if orig.get('output') != repl.get('output'):
                    differences.append({
                        "step_id": step_id,
                        "type": "output_changed",
                        "original_keys": list(orig.get('output', {}).keys()),
                        "replay_keys": list(repl.get('output', {}).keys()),
                    })

        return {
            "original_execution_id": original_execution_id,
            "replay_execution_id": replay_execution_id,
            "original_step_count": len(original_steps),
            "replay_step_count": len(replay_steps),
            "differences": differences,
            "identical": len(differences) == 0,
        }

    async def _load_workflow_definition(
        self,
        execution_id: str,
    ) -> Optional[Dict[str, Any]]:
        """Load workflow definition from execution metadata"""
        meta_path = self.evidence_path / execution_id / "workflow.json"

        if meta_path.exists():
            try:
                with open(meta_path, 'r', encoding='utf-8') as f:
                    return json.load(f)
            except Exception as e:
                logger.warning(f"Failed to load workflow definition: {e}")

        return None

    def _find_step_config(
        self,
        workflow: Optional[Dict[str, Any]],
        step_id: str,
    ) -> Optional[Dict[str, Any]]:
        """Find step configuration by ID"""
        if not workflow:
            return None

        for step in workflow.get('steps', []):
            if step.get('id') == step_id:
                return step

        return None

    def _add_to_history(self, result: ReplayResult) -> None:
        """Add replay result to history"""
        self._replay_history.append(result)

        # Trim history if needed
        if len(self._replay_history) > self.max_replay_history:
            self._replay_history = self._replay_history[-self.max_replay_history:]

    def get_replay_history(
        self,
        execution_id: Optional[str] = None,
    ) -> List[ReplayResult]:
        """Get replay history, optionally filtered by original execution"""
        if execution_id:
            return [
                r for r in self._replay_history
                if r.original_execution_id == execution_id
            ]
        return list(self._replay_history)


# =============================================================================
# Factory Functions
# =============================================================================

def create_replay_manager(
    evidence_path: Optional[Path] = None,
) -> ReplayManager:
    """
    Create a replay manager.

    Args:
        evidence_path: Path to evidence storage

    Returns:
        Configured ReplayManager
    """
    if evidence_path is None:
        evidence_path = Path("./evidence")
    return ReplayManager(evidence_path=evidence_path)



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Composite Modules (Level 3)

High-level workflow templates combining multiple atomic and third-party modules.
Designed for normal users, similar to n8n nodes (3-10 atomic steps).

Categories:
- browser: Web automation workflows
- developer: Developer tool integrations
- notification: Multi-channel notification workflows
- data: Data transformation pipelines

Usage:
    from core.modules.composite import CompositeRegistry, CompositeExecutor

    # List all composites
    composites = CompositeRegistry.list_all()

    # Execute a composite
    executor = CompositeExecutor()
    result = await executor.execute(
        'composite.browser.scrape_to_json',
        {'url': 'https://example.com', 'title_selector': 'h1'}
    )
"""

from .base import (
    CompositeModule,
    CompositeRegistry,
    CompositeExecutor,
    register_composite,
    UIVisibility,
)

# Browser composites
from .browser.scrape_to_json import WebScrapeToJson

# Developer composites
from .developer.github_daily_digest import GithubDailyDigest
from .developer.api_to_notification import ApiToNotification

# Notification composites
from .notification.multi_channel_alert import MultiChannelAlert
from .notification.scheduled_report import ScheduledReport

# Data composites
from .data.csv_to_json import CsvToJson
from .data.json_transform_notify import JsonTransformNotify

# Test composites
from .test.e2e_flow import E2EFlowTest
from .test.api_test import ApiTestSuite
from .test.ui_review import UIReview
from .test.quality_gate import QualityGate


__all__ = [
    # Base classes
    'CompositeModule',
    'CompositeRegistry',
    'CompositeExecutor',
    'register_composite',
    'UIVisibility',

    # Browser composites
    'WebScrapeToJson',

    # Developer composites
    'GithubDailyDigest',
    'ApiToNotification',

    # Notification composites
    'MultiChannelAlert',
    'ScheduledReport',

    # Data composites
    'CsvToJson',
    'JsonTransformNotify',

    # Test composites
    'E2EFlowTest',
    'ApiTestSuite',
    'UIReview',
    'QualityGate',
]


def get_composite_statistics():
    """Get statistics about registered composite modules"""
    return CompositeRegistry.get_statistics()


def list_composites_by_category(category: str):
    """List all composites in a category"""
    return CompositeRegistry.get_all_metadata(category=category)



# Flyto2 Core - Source Available License
# Copyright (c) 2025 Flyto2. All Rights Reserved.
#
# This source code is licensed under the Flyto2 Source Available License v1.0.
# Commercial use requires a license. See LICENSE for details.

"""
Crypto JWT Verify Module
Verify and decode JWT (JSON Web Token) tokens.
"""
import logging
from typing import Any, Dict

from ...registry import register_module
from ...schema import compose
from ...schema.builders import field
from ...schema.constants import FieldGroup
from ...errors import ValidationError, ModuleError


logger = logging.getLogger(__name__)


@register_module(
    module_id='crypto.jwt_verify',
    version='1.0.0',
    category='crypto',
    tags=['crypto', 'jwt', 'token', 'verify', 'auth', 'security', 'advanced'],
    label='Verify JWT',
    label_key='modules.crypto.jwt_verify.label',
    description='Verify and decode JWT tokens',
    description_key='modules.crypto.jwt_verify.description',
    icon='ShieldCheck',
    color='#10B981',
    input_types=['string'],
    output_types=['object'],

    can_receive_from=['*'],
    can_connect_to=['data.*', 'logic.*', 'flow.*'],

    retryable=False,
    concurrent_safe=True,
    timeout_ms=5000,

    requires_credentials=False,
    handles_sensitive_data=True,
    required_permissions=[],

    params_schema=compose(
        field(
            'token',
            type='text',
            label='JWT Token',
            label_key='modules.crypto.jwt_verify.params.token.label',
            description='JWT token to verify and decode',
            description_key='modules.crypto.jwt_verify.params.token.description',
            required=True,
            placeholder='eyJhbGciOiJIUzI1NiIs...',
            group=FieldGroup.BASIC,
        ),
        field(
            'secret',
            type='password',
            label='Secret Key',
            label_key='modules.crypto.jwt_verify.params.secret.label',
            description='Secret key for verifying the token signature',
            description_key='modules.crypto.jwt_verify.params.secret.description',
            required=True,
            placeholder='your-jwt-secret',
            group=FieldGroup.BASIC,
        ),
        field(
            'algorithms',
            type='array',
            label='Allowed Algorithms',
            label_key='modules.crypto.jwt_verify.params.algorithms.label',
            description='List of allowed signing algorithms',
            description_key='modules.crypto.jwt_verify.params.algorithms.description',
            default=['HS256'],
            group=FieldGroup.OPTIONS,
        ),
        field(
            'verify_exp',
            type='boolean',
            label='Verify Expiration',
            label_key='modules.crypto.jwt_verify.params.verify_exp.label',
            description='Whether to verify the token expiration',
            description_key='modules.crypto.jwt_verify.params.verify_exp.description',
            default=True,
            group=FieldGroup.OPTIONS,
        ),
        field(
            'audience',
            type='string',
            label='Audience',
            label_key='modules.crypto.jwt_verify.params.audience.label',
            description='Expected audience (aud claim)',
            description_key='modules.crypto.jwt_verify.params.audience.description',
            required=False,
            placeholder='https://api.example.com',
            group=FieldGroup.OPTIONS,
        ),
        field(
            'issuer',
            type='string',
            label='Issuer',
            label_key='modules.crypto.jwt_verify.params.issuer.label',
            description='Expected issuer (iss claim)',
            description_key='modules.crypto.jwt_verify.params.issuer.description',
            required=False,
            placeholder='https://example.com',
            group=FieldGroup.OPTIONS,
        ),
    ),
    output_schema={
        'valid': {
            'type': 'boolean',
            'description': 'Whether the token is valid',
            'description_key': 'modules.crypto.jwt_verify.output.valid.description',
        },
        'payload': {
            'type': 'object',
            'description': 'Decoded JWT payload',
            'description_key': 'modules.crypto.jwt_verify.output.payload.description',
        },
        'header': {
            'type': 'object',
            'description': 'Decoded JWT header',
            'description_key': 'modules.crypto.jwt_verify.output.header.description',
        },
    },
    examples=[
        {
            'title': 'Verify a JWT token',
            'title_key': 'modules.crypto.jwt_verify.examples.basic.title',
            'params': {
                'token': 'eyJhbGciOiJIUzI1NiIs...',
                'secret': 'my-jwt-secret',
                'algorithms': ['HS256'],
                'verify_exp': True,
            },
        },
    ],
    author='Flyto Team',
    license='MIT',
)
async def crypto_jwt_verify(context: Dict[str, Any]) -> Dict[str, Any]:
    """Verify and decode a JWT token."""
    try:
        import jwt
    except ImportError:
        raise ModuleError(
            "PyJWT is required for crypto.jwt_verify. "
            "Install with: pip install PyJWT"
        )

    params = context['params']
    token = params.get('token')
    secret = params.get('secret')
    algorithms = params.get('algorithms', ['HS256'])
    verify_exp = params.get('verify_exp', True)
    audience = params.get('audience')
    issuer = params.get('issuer')

    if not token:
        raise ValidationError("Missing required parameter: token", field="token")
    if not secret:
        raise ValidationError("Missing required parameter: secret", field="secret")

    # Ensure algorithms is a list
    if isinstance(algorithms, str):
        algorithms = [algorithms]

    # Get header without verification first (for reporting)
    try:
        header = jwt.get_unverified_header(token)
    except jwt.exceptions.DecodeError as e:
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': {},
                'header': {},
                'error': f"Malformed token: {e}",
            },
        }

    # Build decode options
    decode_options = {}
    if not verify_exp:
        decode_options['verify_exp'] = False

    decode_kwargs = {
        'algorithms': algorithms,
        'options': decode_options,
    }
    if audience:
        decode_kwargs['audience'] = audience
    if issuer:
        decode_kwargs['issuer'] = issuer

    try:
        payload = jwt.decode(token, secret, **decode_kwargs)
        logger.info(f"JWT verified successfully (alg={header.get('alg')})")

        return {
            'ok': True,
            'data': {
                'valid': True,
                'payload': payload,
                'header': header,
            },
        }

    except jwt.ExpiredSignatureError:
        # Decode without verification to return payload
        payload = jwt.decode(token, secret, algorithms=algorithms, options={'verify_exp': False})
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': payload,
                'header': header,
                'error': 'Token has expired',
            },
        }

    except jwt.InvalidAudienceError:
        payload = jwt.decode(token, secret, algorithms=algorithms, options={'verify_exp': False, 'verify_aud': False})
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': payload,
                'header': header,
                'error': 'Invalid audience',
            },
        }

    except jwt.InvalidIssuerError:
        payload = jwt.decode(token, secret, algorithms=algorithms, options={'verify_exp': False, 'verify_iss': False})
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': payload,
                'header': header,
                'error': 'Invalid issuer',
            },
        }

    except jwt.InvalidSignatureError:
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': {},
                'header': header,
                'error': 'Invalid signature',
            },
        }

    except jwt.DecodeError as e:
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': {},
                'header': header,
                'error': f"Decode error: {e}",
            },
        }

    except Exception as e:
        return {
            'ok': True,
            'data': {
                'valid': False,
                'payload': {},
                'header': header,
                'error': f"Verification failed: {e}",
            },
        }
